import logging

import pytest

from amqtt.broker import BrokerContext, Broker

from amqtt.contexts import BaseContext, Action
from amqtt.plugins.topic_checking import TopicAccessControlListPlugin, TopicTabooPlugin
from amqtt.plugins.base import BaseTopicPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)

# Base plug-in object


@pytest.mark.asyncio
async def test_base_no_config(logdog):
    """Check BaseTopicPlugin returns false if no topic-check is present."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {}

        plugin = BaseTopicPlugin(context)
        authorised = await plugin.topic_filtering()
        assert authorised is False

        # Warning messages are only generated if using deprecated plugin configuration on initial load
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 1
        assert log_records[0].levelno == logging.WARNING
        assert log_records[0].message == "'topic-check' section not found in context configuration"


@pytest.mark.asyncio
async def test_base_empty_config(logdog):
    """Check BaseTopicPlugin returns false if topic-check is empty."""
    with logdog() as pile:
        broker = Broker()
        context = BrokerContext(broker)
        context.logger = logging.getLogger("testlog")
        context.config = {"topic-check": {}}

        plugin = BaseTopicPlugin(context)
        authorised = await plugin.topic_filtering()
        assert authorised is False

        # Warning messages are only generated if using deprecated plugin configuration on initial load
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 1
        assert log_records[0].levelno == logging.WARNING
        assert log_records[0].message == "'topic-check' section not found in context configuration"


@pytest.mark.asyncio
async def test_base_disabled_config(logdog):
    """Check BaseTopicPlugin returns true if disabled. (it doesn't actually check)."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {"topic-check": {"enabled": False}}

        plugin = BaseTopicPlugin(context)
        authorised = await plugin.topic_filtering()
        assert authorised is True

        # Should NOT have printed warnings
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 0


@pytest.mark.asyncio
async def test_base_enabled_config(logdog):
    """Check BaseTopicPlugin returns true if enabled."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {"topic-check": {"enabled": True}}

        plugin = BaseTopicPlugin(context)
        authorised = await plugin.topic_filtering()
        assert authorised is True

        # Should NOT have printed warnings
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 0


# Taboo plug-in


@pytest.mark.asyncio
async def test_taboo_empty_config(logdog):
    """Check TopicTabooPlugin returns false if topic-check absent."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {}

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering()) is False

        # Warning messages are only generated if using deprecated plugin configuration on initial load
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 1
        assert log_records[0].levelno == logging.WARNING
        assert log_records[0].message == "'topic-check' section not found in context configuration"


@pytest.mark.asyncio
async def test_taboo_disabled(logdog):
    """Check TopicTabooPlugin returns true if checking disabled."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {"topic-check": {"enabled": False}}

        session = Session()
        session.username = "anybody"

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering(session=session, topic="not/prohibited")) is True

    # Should NOT have printed warnings
    log_records = list(pile.drain(name="testlog"))
    assert len(log_records) == 0


@pytest.mark.parametrize("test_config", [
    ({"topic-check": {"enabled": True}}),
    (TopicTabooPlugin.Config())
])
@pytest.mark.asyncio
async def test_taboo_not_taboo_topic(logdog, test_config):
    """Check TopicTabooPlugin returns true if topic not taboo."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = test_config

        session = Session()
        session.username = "anybody"

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering(session=session, topic="not/prohibited")) is True

    # Should NOT have printed warnings
    log_records = list(pile.drain(name="testlog"))
    assert len(log_records) == 0


@pytest.mark.parametrize("test_config", [
    ({"topic-check": {"enabled": True}}),
    (TopicTabooPlugin.Config())
])
@pytest.mark.asyncio
async def test_taboo_anon_taboo_topic(logdog, test_config):
    """Check TopicTabooPlugin returns false if topic is taboo and session is anonymous."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = test_config

        session = Session()
        session.username = ""

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering(session=session, topic="prohibited")) is False

    # Should NOT have printed warnings
    log_records = list(pile.drain(name="testlog"))
    assert len(log_records) == 0


@pytest.mark.parametrize("test_config", [
    ({"topic-check": {"enabled": True}}),
    (TopicTabooPlugin.Config())
])
@pytest.mark.asyncio
async def test_taboo_notadmin_taboo_topic(logdog, test_config):
    """Check TopicTabooPlugin returns false if topic is taboo and user is not "admin"."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = test_config

        session = Session()
        session.username = "notadmin"

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering(session=session, topic="prohibited")) is False

    # Should NOT have printed warnings
    log_records = list(pile.drain(name="testlog"))
    assert len(log_records) == 0

@pytest.mark.parametrize("test_config", [
    ({"topic-check": {"enabled": True}}),
    (TopicTabooPlugin.Config())
])
@pytest.mark.asyncio
async def test_taboo_admin_taboo_topic(logdog, test_config):
    """Check TopicTabooPlugin returns true if topic is taboo and user is "admin"."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = test_config

        session = Session()
        session.username = "admin"

        plugin = TopicTabooPlugin(context)
        assert (await plugin.topic_filtering(session=session, topic="prohibited")) is True

    # Should NOT have printed warnings
    log_records = list(pile.drain(name="testlog"))
    assert len(log_records) == 0


# TopicAccessControlListPlugin tests


def test_topic_ac_not_match():
    """Test TopicAccessControlListPlugin.topic_ac returns false if topics do not match."""
    assert TopicAccessControlListPlugin.topic_ac("a/topic/to/match", "a/topic/to/notmatch") is False


def test_topic_ac_not_match_longer_acl():
    """Test TopicAccessControlListPlugin.topic_ac returns false if topics do not match and ACL topic is longer."""
    assert TopicAccessControlListPlugin.topic_ac("topic", "topic/is/longer") is False


def test_topic_ac_not_match_longer_rq():
    """Test TopicAccessControlListPlugin.topic_ac returns false if topics do not match and RQ topic is longer."""
    assert TopicAccessControlListPlugin.topic_ac("topic/is/longer", "topic") is False


def test_topic_ac_match_exact():
    """Test TopicAccessControlListPlugin.topic_ac returns true if topics match exactly."""
    assert TopicAccessControlListPlugin.topic_ac("exact/topic", "exact/topic") is True


def test_topic_ac_match_plus():
    """Test TopicAccessControlListPlugin.topic_ac correctly handles '+' wildcard."""
    assert (
        TopicAccessControlListPlugin.topic_ac(
            "a/topic/anything/value",
            "a/topic/+/value",
        )
        is True
    )


def test_topic_ac_match_hash():
    """Test TopicAccessControlListPlugin.topic_ac correctly handles '#' wildcard."""
    assert (
        TopicAccessControlListPlugin.topic_ac(
            "topic/prefix/and/suffix",
            "topic/prefix/#",
        )
        is True
    )


@pytest.mark.asyncio
async def test_taclp_empty_config(logdog):
    """Check TopicAccessControlListPlugin returns false if topic-check absent."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {}

        plugin = TopicAccessControlListPlugin(context)
        assert (await plugin.topic_filtering()) is False

        # Warning messages are only generated if using deprecated plugin configuration on initial load
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 1
        assert log_records[0].levelno == logging.WARNING
        assert log_records[0].message == "'topic-check' section not found in context configuration"


@pytest.mark.asyncio
async def test_taclp_true_disabled(logdog):
    """Check TopicAccessControlListPlugin returns true if topic checking is disabled."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = {"topic-check": {"enabled": False}}

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.PUBLISH,
        session=session,
        topic="a/topic",
    )
    assert authorised is True


@pytest.mark.parametrize("test_config", [
    ({"topic-check": {"enabled": True}}),
    (TopicAccessControlListPlugin.Config())
])
@pytest.mark.asyncio
async def test_taclp_true_no_pub_acl(logdog, test_config):
    """Check TopicAccessControlListPlugin returns true if action=publish and no publish-acl given.

    (This is for backward-compatibility with existing installations.).
    """
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.PUBLISH,
        session=session,
        topic="a/topic",
    )
    assert authorised is True


@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "acl": {"anotheruser": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        acl={"anotheruser": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_false_sub_no_topic(logdog, test_config):
    """Check TopicAccessControlListPlugin returns false user there is no topic."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.SUBSCRIBE,
        session=session,
        topic="",
    )
    assert authorised is False


@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "acl": {"anotheruser": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        acl={"anotheruser": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_false_sub_unknown_user(logdog, test_config):
    """Check TopicAccessControlListPlugin returns false user is not listed in ACL."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.SUBSCRIBE,
        session=session,
        topic="allowed/topic",
    )
    assert authorised is False


@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "acl": {"user": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        acl={"user": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_false_sub_no_permission(logdog, test_config):
    """Check TopicAccessControlListPlugin returns false if "acl" does not list allowed topic."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.SUBSCRIBE,
        session=session,
        topic="forbidden/topic",
    )
    assert authorised is False

@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "acl": {"user": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        acl={"user": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_true_sub_permission(logdog, test_config):
    """Check TopicAccessControlListPlugin returns true if "acl" lists allowed topic."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.SUBSCRIBE,
        session=session,
        topic="allowed/topic",
    )
    assert authorised is True


@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "publish-acl": {"user": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        publish_acl={"user": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_true_pub_permission(logdog, test_config):
    """Check TopicAccessControlListPlugin returns true if "publish-acl" lists allowed topic for publish action."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = "user"

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.PUBLISH,
        session=session,
        topic="allowed/topic",
    )
    assert authorised is True


@pytest.mark.parametrize("test_config", [
    ({
        "topic-check": {
            "enabled": True,
            "acl": {"anonymous": ["allowed/topic", "another/allowed/topic/#"]},
        },
    }),
    (TopicAccessControlListPlugin.Config(
        acl={"anonymous": ["allowed/topic", "another/allowed/topic/#"]}
    ))
])
@pytest.mark.asyncio
async def test_taclp_true_anon_sub_permission(logdog, test_config):
    """Check TopicAccessControlListPlugin handles anonymous users."""
    context = BaseContext()
    context.logger = logging.getLogger("testlog")
    context.config = test_config

    session = Session()
    session.username = None

    plugin = TopicAccessControlListPlugin(context)
    authorised = await plugin.topic_filtering(
        action=Action.SUBSCRIBE,
        session=session,
        topic="allowed/topic",
    )
    assert authorised is True



import asyncio
import logging
from pathlib import Path
from typing import Annotated

import click
import passlib
import typer

from amqtt.contrib.auth_db import DBType, db_connection_str
from amqtt.contrib.auth_db.managers import UserManager
from amqtt.errors import MQTTError

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)
user_app = typer.Typer(no_args_is_help=True)


@user_app.callback()
def main(
        ctx: typer.Context,
        db_type: Annotated[DBType, typer.Option(..., "--db", "-d", help="db type", show_default=False)],
        db_username: Annotated[str, typer.Option("--username", "-u", help="db username", show_default=False)] = "",
        db_port: Annotated[int, typer.Option("--port", "-p", help="database port (defaults to db type)", show_default=False)] = 0,
        db_host: Annotated[str, typer.Option("--host", "-h", help="database host")] = "localhost",
        db_filename: Annotated[str, typer.Option("--file", "-f", help="database file name (sqlite only)")] = "auth.db",
) -> None:
    """Command line interface to list, create, remove and add clients.

    Passwords are not allowed to be passed via the command line for security reasons. You will be prompted for database
    password (if applicable) and the client id's password.

    If you need to create users programmatically, see `amqtt.contrib.auth_db.managers.UserManager` which provides
    the underlying functionality to this command line interface.
    """
    if db_type == DBType.SQLITE and ctx.invoked_subcommand == "sync" and not Path(db_filename).exists():
        pass
    elif db_type == DBType.SQLITE and not Path(db_filename).exists():
        logger.error(f"SQLite option could not find '{db_filename}'")
        raise typer.Exit(code=1)
    elif db_type != DBType.SQLITE and not db_username:
        logger.error("DB access requires a username be provided.")
        raise typer.Exit(code=1)

    ctx.obj = {"type": db_type, "username": db_username, "host": db_host, "port": db_port, "filename": db_filename}


@user_app.command(name="sync")
def db_sync(ctx: typer.Context) -> None:
    """Create the table and schema for username and hashed password.

    Non-destructive if run multiple times. To clear the whole table, need to drop it manually.
    """
    async def run_sync() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"], ctx.obj["filename"])
        mgr = UserManager(connect)
        try:
            await mgr.db_sync()
        except MQTTError as me:
            logger.critical("Could not sync schema on db.")
            raise typer.Exit(code=1) from me

    asyncio.run(run_sync())
    logger.info("Success: database synced.")


@user_app.command(name="list")
def list_user_auths(ctx: typer.Context) -> None:
    """List all Client IDs (in alphabetical order). Will also display the hashed passwords."""

    async def run_list() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"], ctx.obj["filename"])
        mgr = UserManager(connect)
        user_count = 0
        for user in await mgr.list_user_auths():
            user_count += 1
            logger.info(user)

        if not user_count:
            logger.info("No client authentications exist.")

    asyncio.run(run_list())


@user_app.command(name="add")
def create_user_auth(
        ctx: typer.Context,
        client_id: Annotated[str, typer.Option("--client-id", "-c", help="id for the new client")],
        ) -> None:
    """Create a new user with a client id and password (prompted)."""
    async def run_create() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"],
                                    ctx.obj["filename"])
        mgr = UserManager(connect)
        client_password = click.prompt("Enter the client's password", hide_input=True)
        if not client_password.strip():
            logger.info("Error: client password cannot be empty.")
            raise typer.Exit(1)
        try:
            user = await mgr.create_user_auth(client_id, client_password.strip())
        except passlib.exc.MissingBackendError as mbe:
            logger.info(f"Please install backend: {mbe}")
            raise typer.Exit(code=1) from mbe

        if not user:
            logger.info(f"Error: could not create user: {client_id}")
            raise typer.Exit(code=1)

        logger.info(f"Success: created {user}")

    asyncio.run(run_create())


@user_app.command(name="rm")
def remove_user_auth(ctx: typer.Context,
                     client_id: Annotated[str, typer.Option("--client-id", "-c", help="id for the client to remove")]) -> None:
    """Remove a client from the authentication database."""
    async def run_remove() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"],
                                    ctx.obj["filename"])
        mgr = UserManager(connect)
        user = await mgr.get_user_auth(client_id)
        if not user:
            logger.info(f"Error: client '{client_id}' does not exist.")
            raise typer.Exit(1)

        if not click.confirm(f"Please confirm the removal of '{client_id}'?"):
            raise typer.Exit(0)

        user = await mgr.delete_user_auth(client_id)
        if not user:
            logger.info(f"Error: client '{client_id}' does not exist.")
            raise typer.Exit(1)

        logger.info(f"Success: '{user.username}' was removed.")

    asyncio.run(run_remove())


@user_app.command(name="pwd")
def change_password(
        ctx: typer.Context,
        client_id: Annotated[str, typer.Option("--client-id", "-c", help="id for the new client")],
        ) -> None:
    """Update a user's password (prompted)."""
    async def run_password() -> None:
        client_password = click.prompt("Enter the client's new password", hide_input=True)
        if not client_password.strip():
            logger.error("Error: client password cannot be empty.")
            raise typer.Exit(1)
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"],
                                    ctx.obj["filename"])
        mgr = UserManager(connect)
        await mgr.update_user_auth_password(client_id, client_password.strip())
        logger.info(f"Success: client '{client_id}' password updated.")

    asyncio.run(run_password())


if __name__ == "__main__":
    user_app()



import asyncio
import contextlib
import logging
from pathlib import Path
from typing import Annotated

import typer

from amqtt.contexts import Action
from amqtt.contrib.auth_db import DBType, db_connection_str
from amqtt.contrib.auth_db.managers import TopicManager, UserManager
from amqtt.errors import MQTTError

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)
topic_app = typer.Typer(no_args_is_help=True)


@topic_app.callback()
def main(
        ctx: typer.Context,
        db_type: Annotated[DBType, typer.Option("--db", "-d", help="db type", count=False)],
        db_username: Annotated[str, typer.Option("--username", "-u", help="db username", show_default=False)] = "",
        db_port: Annotated[int, typer.Option("--port", "-p", help="database port (defaults to db type)", show_default=False)] = 0,
        db_host: Annotated[str, typer.Option("--host", "-h", help="database host")] = "localhost",
        db_filename: Annotated[str, typer.Option("--file", "-f", help="database file name (sqlite only)")] = "auth.db",
) -> None:
    """Command line interface to add / remove topic authorization.

    Passwords are not allowed to be passed via the command line for security reasons. You will be prompted for database
    password (if applicable).

    If you need to create users programmatically, see `amqtt.contrib.auth_db.managers.TopicManager` which provides
    the underlying functionality to this command line interface.
    """
    if db_type == DBType.SQLITE and ctx.invoked_subcommand == "sync" and not Path(db_filename).exists():
        pass
    elif db_type == DBType.SQLITE and not Path(db_filename).exists():
        logger.error(f"SQLite option could not find '{db_filename}'")
        raise typer.Exit(code=1)
    elif db_type != DBType.SQLITE and not db_username:
        logger.error("DB access requires a username be provided.")
        raise typer.Exit(code=1)

    ctx.obj = {"type": db_type, "username": db_username, "host": db_host, "port": db_port, "filename": db_filename}


@topic_app.command(name="sync")
def db_sync(ctx: typer.Context) -> None:
    """Create the table and schema for username and topic lists for subscribe, publish or receive.

    Non-destructive if run multiple times. To clear the whole table, need to drop it manually.
    """
    async def run_sync() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"], ctx.obj["filename"])
        mgr = UserManager(connect)
        try:
            await mgr.db_sync()
        except MQTTError as me:
            logger.critical("Could not sync schema on db.")
            raise typer.Exit(code=1) from me
    asyncio.run(run_sync())
    logger.info("Success: database synced.")


@topic_app.command(name="list")
def list_clients(ctx: typer.Context) -> None:
    """List all Client IDs (in alphabetical order). Will also display the hashed passwords."""

    async def run_list() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"], ctx.obj["filename"])
        mgr = TopicManager(connect)
        user_count = 0
        for user in await mgr.list_topic_auths():
            user_count += 1
            logger.info(user)

        if not user_count:
            logger.info("No client authorizations exist.")

    asyncio.run(run_list())


@topic_app.command(name="add")
def add_topic_allowance(
        ctx: typer.Context,
        topic: Annotated[str, typer.Argument(help="list of topics", show_default=False)],
        client_id: Annotated[str, typer.Option("--client-id", "-c", help="id for the client", show_default=False)],
        action: Annotated[Action, typer.Option("--action", "-a", help="action for topic to allow", show_default=False)]
        ) -> None:
    """Create a new user with a client id and password (prompted)."""
    async def run_add() -> None:

        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"],
                                    ctx.obj["filename"])
        mgr = TopicManager(connect)

        with contextlib.suppress(MQTTError):
            await mgr.create_topic_auth(client_id)

        topic_auth = await mgr.get_topic_auth(client_id)
        if not topic_auth:
            logger.info(f"Topic auth doesn't exist for '{client_id}'")
            raise typer.Exit(code=1)

        if topic in [allowed_topic.topic for allowed_topic in topic_auth.get_topic_list(action)]:
            logger.info(f"Topic '{topic}' already exists for '{action}'.")
            raise typer.Exit(1)

        await mgr.add_allowed_topic(client_id, topic, action)

        logger.info(f"Success: topic '{topic}' added to {action} for '{client_id}'")

    asyncio.run(run_add())


@topic_app.command(name="rm")
def remove_topic_allowance(ctx: typer.Context,
                           client_id: Annotated[str, typer.Option("--client-id", "-c", help="id for the client to remove")],
                           action: Annotated[Action, typer.Option("--action", "-a", help="action for topic to allow")],
                           topic: Annotated[str, typer.Argument(help="list of topics")]
                           ) -> None:
    """Remove a client from the authentication database."""
    async def run_remove() -> None:
        connect = db_connection_str(ctx.obj["type"], ctx.obj["username"], ctx.obj["host"], ctx.obj["port"],
                                    ctx.obj["filename"])
        mgr = TopicManager(connect)

        topic_auth = await mgr.get_topic_auth(client_id)

        if not topic_auth:
            logger.info(f"client '{client_id}' doesn't exist.")
            raise typer.Exit(1)

        if topic not in getattr(topic_auth, f"{action}_acl"):
            logger.info(f"Error: topic '{topic}' not in the {action} allow list for {client_id}.")
            raise typer.Exit(1)

        try:
            await mgr.remove_allowed_topic(client_id, topic, action)
        except MQTTError as me:
            logger.info(f"'Error: could not remove '{topic}' for client '{client_id}'.")
            raise typer.Exit(1) from me

        logger.info(f"Success: removed topic '{topic}' from {action} for '{client_id}'")

    asyncio.run(run_remove())


if __name__ == "__main__":
    topic_app()



import asyncio
import unittest
import pytest

from amqtt.adapters import BufferReader
from amqtt.errors import AMQTTError
from amqtt.mqtt.connect import ConnectPacket, ConnectPayload, ConnectVariableHeader
from amqtt.mqtt.packet import CONNECT, MQTTFixedHeader, PUBLISH


class ConnectPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_decode_ok(self):
        data = (
            b"\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789"
            b"\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        )
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.variable_header is not None
        assert message.variable_header.proto_name == "MQTT"
        assert message.variable_header.proto_level == 4
        assert message.variable_header.username_flag
        assert message.variable_header.password_flag
        assert not message.variable_header.will_retain_flag
        assert message.variable_header.will_qos == 1
        assert message.variable_header.will_flag
        assert message.variable_header.clean_session_flag
        assert not message.variable_header.reserved_flag

        assert message.payload is not None
        assert message.payload.client_id == "0123456789"
        assert message.payload.will_topic == "WillTopic"
        assert message.payload.will_message == b"WillMessage"
        assert message.payload.username == "user"
        assert message.payload.password == "password"

    def test_decode_ok_will_flag(self):
        data = b"\x10\x26\x00\x04MQTT\x04\xca\x00\x00\x00\x0a0123456789\x00\x04user\x00\x08password"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.variable_header is not None
        assert message.variable_header.proto_name == "MQTT"
        assert message.variable_header.proto_level == 4
        assert message.variable_header.username_flag
        assert message.variable_header.password_flag
        assert not message.variable_header.will_retain_flag
        assert message.variable_header.will_qos == 1
        assert not message.variable_header.will_flag
        assert message.variable_header.clean_session_flag
        assert not message.variable_header.reserved_flag

        assert message.payload is not None
        assert message.payload.client_id == "0123456789"
        assert message.payload.will_topic is None
        assert message.payload.will_message is None
        assert message.payload.username == "user"
        assert message.payload.password == "password"

    def test_decode_fail_reserved_flag(self):
        data = (
            b"\x10\x3e\x00\x04MQTT\x04\xcf\x00\x00\x00\x0a0123456789"
            b"\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        )
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.variable_header is not None
        assert message.variable_header.reserved_flag

    def test_decode_fail_miss_clientId(self):
        data = b"\x10\x0a\x00\x04MQTT\x04\xce\x00\x00"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.payload is not None
        assert message.payload.client_id is not None

    def test_decode_fail_miss_willtopic(self):
        data = b"\x10\x16\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.payload is not None
        assert message.payload.will_topic is None

    def test_decode_fail_miss_username(self):
        data = b"\x10\x2e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.payload is not None
        assert message.payload.username is None

    def test_decode_fail_miss_password(self):
        data = b"\x10\x34\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.payload is not None
        assert message.payload.password is None

    def test_encode(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 0)
        variable_header = ConnectVariableHeader(0xCE, 0, "MQTT", 4)
        payload = ConnectPayload(
            "0123456789",
            "WillTopic",
            b"WillMessage",
            "user",
            "password",
        )
        message = ConnectPacket(header, variable_header, payload)
        encoded = message.to_bytes()
        assert (
            encoded
            == b"\x10>\x00\x04MQTT\x04\xce\x00\x00\x00\n0123456789\x00\tWillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        )

    def test_getattr_ok(self):
        data = (
            b"\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789"
            b"\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password"
        )
        stream = BufferReader(data)
        message = self.loop.run_until_complete(ConnectPacket.from_stream(stream))
        assert message.variable_header is not None
        assert message.variable_header.proto_name == "MQTT"
        assert message.proto_name == "MQTT"
        assert message.variable_header.proto_level == 4
        assert message.proto_level == 4
        assert message.variable_header.username_flag
        assert message.username_flag
        assert message.variable_header.password_flag
        assert message.password_flag
        assert not message.variable_header.will_retain_flag
        assert not message.will_retain_flag
        assert message.variable_header.will_qos == 1
        assert message.will_qos == 1
        assert message.variable_header.will_flag
        assert message.will_flag
        assert message.variable_header.clean_session_flag
        assert message.clean_session_flag
        assert not message.variable_header.reserved_flag
        assert not message.reserved_flag

        assert message.payload is not None
        assert message.payload.client_id == "0123456789"
        assert message.client_id == "0123456789"
        assert message.payload.will_topic == "WillTopic"
        assert message.will_topic == "WillTopic"
        assert message.payload.will_message == b"WillMessage"
        assert message.will_message == b"WillMessage"
        assert message.payload.username == "user"
        assert message.username == "user"
        assert message.payload.password == "password"
        assert message.password == "password"

def test_incorrect_fixed_header():
    header = MQTTFixedHeader(PUBLISH, 0x00)
    with pytest.raises(AMQTTError):
        _ = ConnectPacket(fixed=header)

@pytest.mark.parametrize("prop", [
    "proto_name",
    "proto_level",
    "username_flag",
    "password_flag",
    "clean_session_flag",
    "will_retain_flag",
    "will_qos",
    "will_flag",
    "reserved_flag",
    "client_id",
    "client_id_is_random",
    "will_topic",
    "will_message",
    "username",
    "password",
    "keep_alive",
])
def test_empty_variable_header(prop):
    packet = ConnectPacket()

    with pytest.raises(ValueError):
        assert getattr(packet, prop) is not None

    with pytest.raises(ValueError):
        assert setattr(packet, prop, "a value")




import warnings

from amqtt.broker import BrokerContext
from amqtt.plugins.base import BasePlugin


class SQLitePlugin(BasePlugin[BrokerContext]):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        warnings.warn("SQLitePlugin is deprecated, use amqtt.contrib.persistence.SessionDBPlugin", stacklevel=1)



from collections.abc import Iterator
import logging

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

from amqtt.contexts import Action
from amqtt.contrib.auth_db.models import AllowedTopic, Base, TopicAuth, UserAuth
from amqtt.errors import MQTTError

logger = logging.getLogger(__name__)


class UserManager:

    def __init__(self, connection: str) -> None:
        self._engine = create_async_engine(connection)
        self._db_session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    async def db_sync(self) -> None:
        """Sync the database schema."""
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    @staticmethod
    async def _get_auth_or_raise(db_session: AsyncSession, username: str) -> UserAuth:
        stmt = select(UserAuth).filter(UserAuth.username == username)
        user_auth = await db_session.scalar(stmt)
        if not user_auth:
            msg = f"Username '{username}' doesn't exist."
            logger.debug(msg)
            raise MQTTError(msg)

        return user_auth

    async def get_user_auth(self, username: str) -> UserAuth | None:
        """Retrieve a user by username."""
        async with self._db_session_maker() as db_session, db_session.begin():
            try:
                return await self._get_auth_or_raise(db_session, username)
            except MQTTError:
                return None

    async def list_user_auths(self) -> Iterator[UserAuth]:
        """Return list of all clients."""
        async with self._db_session_maker() as db_session, db_session.begin():
            stmt = select(UserAuth).order_by(UserAuth.username)
            users = await db_session.scalars(stmt)
            if not users:
                msg = "No users exist."
                logger.info(msg)
                raise MQTTError(msg)
            return users

    async def create_user_auth(self, username: str, plain_password: str) -> UserAuth | None:
        """Create a new user."""
        async with self._db_session_maker() as db_session, db_session.begin():
            stmt = select(UserAuth).filter(UserAuth.username == username)
            user_auth = await db_session.scalar(stmt)
            if user_auth:
                msg = f"Username '{username}' already exists."
                logger.info(msg)
                raise MQTTError(msg)

            user_auth = UserAuth(username=username)
            user_auth.password = plain_password

            db_session.add(user_auth)
            await db_session.commit()
            await db_session.flush()
            return user_auth

    async def delete_user_auth(self, username: str) -> UserAuth | None:
        """Delete a user."""
        async with self._db_session_maker() as db_session, db_session.begin():

            try:
                user_auth = await self._get_auth_or_raise(db_session, username)
            except MQTTError:
                return None

            await db_session.delete(user_auth)
            await db_session.commit()
            await db_session.flush()
            return user_auth

    async def update_user_auth_password(self, username: str, plain_password: str) -> UserAuth | None:
        """Change a user's password."""
        async with self._db_session_maker() as db_session, db_session.begin():
            user_auth = await self._get_auth_or_raise(db_session, username)
            user_auth.password = plain_password
            await db_session.commit()
            await db_session.flush()
            return user_auth


class TopicManager:

    def __init__(self, connection: str) -> None:
        self._engine = create_async_engine(connection)
        self._db_session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    async def db_sync(self) -> None:
        """Sync the database schema."""
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    @staticmethod
    async def _get_auth_or_raise(db_session: AsyncSession, username: str) -> TopicAuth:
        stmt = select(TopicAuth).filter(TopicAuth.username == username)
        topic_auth = await db_session.scalar(stmt)
        if not topic_auth:
            msg = f"Username '{username}' doesn't exist."
            logger.debug(msg)
            raise MQTTError(msg)

        return topic_auth

    @staticmethod
    def _field_name(action: Action) -> str:
        return f"{action}_acl"

    async def create_topic_auth(self, username: str) -> TopicAuth | None:
        """Create a new user."""
        async with self._db_session_maker() as db_session, db_session.begin():
            stmt = select(TopicAuth).filter(TopicAuth.username == username)
            topic_auth = await db_session.scalar(stmt)
            if topic_auth:
                msg = f"Username '{username}' already exists."
                raise MQTTError(msg)

            topic_auth = TopicAuth(username=username)

            db_session.add(topic_auth)
            await db_session.commit()
            await db_session.flush()
            return topic_auth

    async def get_topic_auth(self, username: str) -> TopicAuth | None:
        """Retrieve a allowed topics by username."""
        async with self._db_session_maker() as db_session, db_session.begin():
            try:
                return await self._get_auth_or_raise(db_session, username)
            except MQTTError:
                return None

    async def list_topic_auths(self) -> Iterator[TopicAuth]:
        """Return list of all authorized clients."""
        async with self._db_session_maker() as db_session, db_session.begin():
            stmt = select(TopicAuth).order_by(TopicAuth.username)
            topics = await db_session.scalars(stmt)
            if not topics:
                msg = "No topics exist."
                logger.info(msg)
                raise MQTTError(msg)
            return topics

    async def add_allowed_topic(self, username: str, topic: str, action: Action) -> list[AllowedTopic] | None:
        """Add allowed topic from action for user."""
        if action == Action.PUBLISH and topic.startswith("$"):
            msg = "MQTT does not allow clients to publish to $ topics."
            raise MQTTError(msg)

        async with self._db_session_maker() as db_session, db_session.begin():
            user_auth = await self._get_auth_or_raise(db_session, username)
            topic_list = getattr(user_auth, self._field_name(action))

            updated_list = [*topic_list, AllowedTopic(topic)]
            setattr(user_auth, self._field_name(action), updated_list)
            await db_session.commit()
            await db_session.flush()
            return updated_list

    async def remove_allowed_topic(self, username: str, topic: str, action: Action) -> list[AllowedTopic] | None:
        """Remove topic from action for user."""
        async with self._db_session_maker() as db_session, db_session.begin():
            topic_auth = await self._get_auth_or_raise(db_session, username)
            topic_list = topic_auth.get_topic_list(action)

            if AllowedTopic(topic) not in topic_list:
                msg = f"Client '{username}' doesn't have topic '{topic}' for action '{action}'."
                logger.debug(msg)
                raise MQTTError(msg)

            updated_list = [allowed_topic for allowed_topic in topic_list if allowed_topic != AllowedTopic(topic)]

            setattr(topic_auth, f"{action}_acl", updated_list)
            await db_session.commit()
            await db_session.flush()
            return updated_list



from collections import defaultdict
from dataclasses import dataclass, field
import json
import re
from typing import Any

from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine

from amqtt.broker import BrokerContext
from amqtt.contexts import Action
from amqtt.contrib.shadows.messages import (
    GetAcceptedMessage,
    GetRejectedMessage,
    UpdateAcceptedMessage,
    UpdateDeltaMessage,
    UpdateDocumentMessage,
    UpdateIotaMessage,
)
from amqtt.contrib.shadows.models import Shadow, sync_shadow_base
from amqtt.contrib.shadows.states import (
    ShadowOperation,
    StateDocument,
    calculate_delta_update,
    calculate_iota_update,
)
from amqtt.plugins.base import BasePlugin, BaseTopicPlugin
from amqtt.session import ApplicationMessage, Session

shadow_topic_re = re.compile(r"^\$shadow/(?P<client_id>[a-zA-Z0-9_-]+?)/(?P<shadow_name>[a-zA-Z0-9_-]+?)/(?P<request>get|update)")

DeviceID = str
ShadowName = str


@dataclass
class ShadowTopic:
    device_id: DeviceID
    name: ShadowName
    message_op: ShadowOperation


def shadow_dict() -> dict[DeviceID, dict[ShadowName, StateDocument]]:
    """Nested defaultdict for shadow cache."""
    return defaultdict(shadow_dict)  # type: ignore[arg-type]


class ShadowPlugin(BasePlugin[BrokerContext]):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        self._shadows: dict[DeviceID, dict[ShadowName, StateDocument]] = defaultdict(dict)

        self._engine = create_async_engine(self.config.connection)
        self._db_session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    async def on_broker_pre_start(self) -> None:
        """Sync the schema."""
        async with self._engine.begin() as conn:
            await sync_shadow_base(conn)

    @staticmethod
    def shadow_topic_match(topic: str) -> ShadowTopic | None:
        """Check if topic matches the shadow topic format."""
        # pattern is "$shadow/<username>/<shadow_name>/get, update, etc
        match = shadow_topic_re.search(topic)
        if match:
            groups = match.groupdict()
            return ShadowTopic(groups["client_id"], groups["shadow_name"], ShadowOperation(groups["request"]))
        return None

    async def _handle_get(self, st: ShadowTopic) -> None:
        """Send 'accepted."""
        async with self._db_session_maker() as db_session, db_session.begin():
            shadow = await Shadow.latest_version(db_session, st.device_id, st.name)
            if not shadow:
                reject_msg = GetRejectedMessage(
                    code=404,
                    message="shadow not found",
                )
                await self.context.broadcast_message(reject_msg.topic(st.device_id, st.name), reject_msg.to_message())
                return

            accept_msg = GetAcceptedMessage(
                state=shadow.state.state,
                metadata=shadow.state.metadata,
                timestamp=shadow.created_at,
                version=shadow.version
            )
            await self.context.broadcast_message(accept_msg.topic(st.device_id, st.name), accept_msg.to_message())

    async def _handle_update(self, st: ShadowTopic, update: dict[str, Any]) -> None:
        async with self._db_session_maker() as db_session, db_session.begin():
            shadow = await Shadow.latest_version(db_session, st.device_id, st.name)
            if not shadow:
                shadow = Shadow(device_id=st.device_id, name=st.name)

            state_update = StateDocument.from_dict(update)

            prev_state = shadow.state or StateDocument()
            prev_state.version = shadow.version or 0  # only required when generating shadow messages
            prev_state.timestamp = shadow.created_at or 0  # only required when generating shadow messages

            next_state = prev_state + state_update

            shadow.state = next_state
            db_session.add(shadow)
            await db_session.commit()

            next_state.version = shadow.version
            next_state.timestamp = shadow.created_at

            accept_msg = UpdateAcceptedMessage(
                state=next_state.state,
                metadata=next_state.metadata,
                timestamp=123,
                version=1
            )

            await self.context.broadcast_message(accept_msg.topic(st.device_id, st.name), accept_msg.to_message())

            delta_msg = UpdateDeltaMessage(
                state=calculate_delta_update(next_state.state.desired, next_state.state.reported),
                metadata=calculate_delta_update(next_state.metadata.desired, next_state.metadata.reported),
                version=shadow.version,
                timestamp=shadow.created_at
            )
            await self.context.broadcast_message(delta_msg.topic(st.device_id, st.name), delta_msg.to_message())

            iota_msg = UpdateIotaMessage(
                state=calculate_iota_update(next_state.state.desired, next_state.state.reported),
                metadata=calculate_delta_update(next_state.metadata.desired, next_state.metadata.reported),
                version=shadow.version,
                timestamp=shadow.created_at
            )
            await self.context.broadcast_message(iota_msg.topic(st.device_id, st.name), iota_msg.to_message())

            doc_msg = UpdateDocumentMessage(
                previous=prev_state,
                current=next_state,
                timestamp=shadow.created_at
            )

            await self.context.broadcast_message(doc_msg.topic(st.device_id, st.name), doc_msg.to_message())

    async def on_broker_message_received(self, *, client_id: str, message: ApplicationMessage) -> None:
        """Process a message that was received from a client."""
        topic = message.topic
        if not topic.startswith("$shadow"):  # this is less overhead than do the full regular expression match
            return

        if not (shadow_topic := self.shadow_topic_match(topic)):
            return

        match shadow_topic.message_op:

            case ShadowOperation.GET:
                await self._handle_get(shadow_topic)
            case ShadowOperation.UPDATE:
                await self._handle_update(shadow_topic, json.loads(message.data.decode("utf-8")))

    @dataclass
    class Config:
        """Configuration for shadow plugin."""

        connection: str
        """SQLAlchemy connection string for the asyncio version of the database connector:

        - `mysql+aiomysql://user:password@host:port/dbname`
        - `postgresql+asyncpg://user:password@host:port/dbname`
        - `sqlite+aiosqlite:///dbfilename.db`
        """


class ShadowTopicAuthPlugin(BaseTopicPlugin):

    async def topic_filtering(self, *,
                        session: Session | None = None,
                        topic: str | None = None,
                        action: Action | None = None) -> bool | None:

        session = session or Session()
        if not topic:
            return False

        shadow_topic = ShadowPlugin.shadow_topic_match(topic)

        if not shadow_topic:
            return False

        return shadow_topic.device_id == session.username or session.username in self.config.superusers

    @dataclass
    class Config:
        """Configuration for only allowing devices access to their own shadow topics."""

        superusers: list[str] = field(default_factory=list)
        """A list of one or more usernames that can write to any device topic,
         primarily for the central app sending updates to devices."""



import asyncio
from asyncio import StreamWriter, StreamReader, Event
from functools import partial
import logging
from pathlib import Path
import ssl

import typer

from amqtt.adapters import ReaderAdapter, WriterAdapter
from amqtt.broker import Broker
from amqtt.client import ClientContext
from amqtt.contexts import BrokerConfig, ClientConfig, ListenerConfig, ListenerType
from amqtt.mqtt.protocol.client_handler import ClientProtocolHandler
from amqtt.plugins.manager import PluginManager
from amqtt.session import Session

logger = logging.getLogger(__name__)


app = typer.Typer(add_completion=False, rich_markup_mode=None)

#   Usage: unix_sockets.py [OPTIONS] COMMAND [ARGS]...
#
#   Options:
#     --help  Show this message and exit.
#
#  Commands:
#     broker  Run an mqtt broker that communicates over a unix (file) socket.
#     client  Run an mqtt client that communicates over a unix (file) socket.


class UnixStreamReaderAdapter(ReaderAdapter):

    def __init__(self, reader: StreamReader) -> None:
        self._reader = reader

    async def read(self, n:int = -1) -> bytes:
        if n < 0:
            return await self._reader.read()
        return await self._reader.readexactly(n)

    def feed_eof(self) -> None:
        return self._reader.feed_eof()


class UnixStreamWriterAdapter(WriterAdapter):

    def __init__(self, writer: StreamWriter) -> None:
        self._writer = writer
        self.is_closed = Event()

    def write(self, data: bytes) -> None:
        if not self.is_closed.is_set():
            self._writer.write(data)

    async def drain(self) -> None:
        if self.is_closed.is_set():
            await self._writer.drain()

    def get_peer_info(self) -> tuple[str, int]:
        extra_info = self._writer.get_extra_info("socket")
        return extra_info.getsockname(), 0

    async def close(self) -> None:
        if self.is_closed.is_set():
            return
        self.is_closed.set()

        await self._writer.drain()
        if self._writer.can_write_eof():
            self._writer.write_eof()

        self._writer.close()

        with contextlib.suppress(AttributeError):
            await self._writer.wait_closed()

    def get_ssl_info(self) -> ssl.SSLObject | None:
        pass


async def run_broker(socket_file: Path):

    # configure the broker with a single, external listener
    cfg = BrokerConfig(
        listeners={
            "default": ListenerConfig(
                type=ListenerType.EXTERNAL
            )
        },
        plugins={
            "amqtt.plugins.logging_amqtt.EventLoggerPlugin":{},
            "amqtt.plugins.logging_amqtt.PacketLoggerPlugin":{},
            "amqtt.plugins.authentication.AnonymousAuthPlugin":{"allow_anonymous":True},
        }
    )

    b = Broker(cfg)

    # new connection handler
    async def unix_stream_connected(reader: StreamReader, writer: StreamWriter, listener_name: str):
        logger.info("received new unix connection....")
        # wraps the reader/writer in a compatible interface
        r = UnixStreamReaderAdapter(reader)
        w = UnixStreamWriterAdapter(writer)

        # passes the connection to the broker for protocol communications
        await b.external_connected(reader=r, writer=w, listener_name=listener_name)

    await asyncio.start_unix_server(partial(unix_stream_connected, listener_name="default"), path=socket_file)
    await b.start()

    try:
        logger.info("starting mqtt unix server")
        # run until ctrl-c
        while True:
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        await b.shutdown()


@app.command()
def broker(
        socket_file: str | None = typer.Option("/tmp/mqtt", "-s", "--socket", help="path and file for unix socket"),
        verbose: bool = typer.Option(False, "-v", "--verbose", help="set logging level to DEBUG"),
):
    """Run an mqtt broker that communicates over a unix (file) socket."""
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    asyncio.run(run_broker(Path(socket_file)))


async def run_client(socket_file: Path):
    # 'MQTTClient' establishes the connection but uses the ClientProtocolHandler for MQTT protocol communications

    # create a plugin manager
    config = ClientConfig()
    context = ClientContext()
    context.config = config
    plugins_manager = PluginManager("amqtt.client.plugins", context)

    # create a client protocol handler
    cph = ClientProtocolHandler(plugins_manager)

    # connect to the unix socket
    conn_reader, conn_writer = await asyncio.open_unix_connection(path=socket_file)

    # anonymous session connection just needs a client_id
    s = Session()
    s.client_id = "myUnixClientID"

    # wraps the reader/writer in compatible interface
    r = UnixStreamReaderAdapter(conn_reader)
    w = UnixStreamWriterAdapter(conn_writer)

    # pass the connection to the protocol handler for mqtt communications and initiate CONNECT/CONNACK
    cph.attach(session=s, reader=r, writer=w)
    logger.debug("handler attached")
    ret = await cph.mqtt_connect()
    logger.info(f"client connected: {ret}")

    try:
        while True:
            # periodically send a message
            await cph.mqtt_publish("my/topic", b"my message", 0, False)
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        cph.detach()


@app.command()
def client(
    socket_file: str | None = typer.Option("/tmp/mqtt", "-s", "--socket", help="path and file for unix socket"),
        verbose: bool = typer.Option(False, "-v", "--verbose", help="set logging level to DEBUG"),
):
    """Run an mqtt client that communicates over a unix (file) socket."""
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    asyncio.run(run_client(Path(socket_file)))


if __name__ == "__main__":
    app()



import asyncio
import unittest

from amqtt.adapters import StreamReaderAdapter
from amqtt.codecs_amqtt import (
    bytes_to_hex_str,
    bytes_to_int,
    decode_string,
    encode_string,
)


class TestCodecs(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_bytes_to_hex_str(self):
        ret = bytes_to_hex_str(b"\x7f")
        assert ret == "0x7f"

    def test_bytes_to_int(self):
        ret = bytes_to_int(b"\x7f")
        assert ret == 127
        ret = bytes_to_int(b"\xff\xff")
        assert ret == 65535

    def test_decode_string(self):
        stream = asyncio.StreamReader(loop=self.loop)
        stream.feed_data(b"\x00\x02AA")
        ret = self.loop.run_until_complete(decode_string(StreamReaderAdapter(stream)))
        assert ret == "AA"

    def test_encode_string(self):
        encoded = encode_string("AA")
        assert encoded == b"\x00\x02AA"



import asyncio
import logging

import pytest

from amqtt.broker import Broker
from amqtt.client import MQTTClient
from amqtt.mqtt.constants import QOS_0


logger = logging.getLogger(__name__)

@pytest.mark.asyncio
async def test_publish_to_dollar_sign_topics():
    """Applications cannot use a topic with a leading $ character for their own purposes [MQTT-4.7.2-1]."""

    cfg = {
        'listeners': {'default': {'type': 'tcp', 'bind': '127.0.0.1'}},
        'plugins': {'amqtt.plugins.authentication.AnonymousAuthPlugin': {"allow_anonymous": True}},
    }

    b = Broker(config=cfg)
    await b.start()
    await asyncio.sleep(0.1)
    c = MQTTClient(config={'auto_reconnect': False})
    await c.connect()
    await asyncio.sleep(0.1)
    await c.subscribe(
        [('$#', QOS_0),
         ('#', QOS_0)]
    )
    await asyncio.sleep(0.1)
    await c.publish('$MY', b'message should be blocked')
    await asyncio.sleep(0.1)

    with pytest.raises(asyncio.TimeoutError):
        # wait long enough for broker sys plugin to run
        _ = await c.deliver_message(timeout_duration=1)

    await c.disconnect()
    await asyncio.sleep(0.1)
    await b.shutdown()

@pytest.mark.asyncio
async def test_hash_will_not_receive_dollar():
    """A subscription to “#” will not receive any messages published to a topic beginning with a $ [MQTT-4.7.2-1]."""

    cfg = {
        'listeners': {'default': {'type': 'tcp', 'bind': '127.0.0.1'}},
        'plugins': {
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {"allow_anonymous": True},
            'amqtt.plugins.sys.broker.BrokerSysPlugin': {"sys_interval": 2}
        }
    }

    b = Broker(config=cfg)
    await b.start()
    await asyncio.sleep(0.1)
    c = MQTTClient(config={'auto_reconnect': False})
    await c.connect()
    await asyncio.sleep(0.1)
    await c.subscribe(
        [('#', QOS_0)]
    )
    await asyncio.sleep(0.1)

    with pytest.raises(asyncio.TimeoutError):
        # wait long enough for broker sys plugin to run
        _ = await c.deliver_message(timeout_duration=5)

    await c.disconnect()
    await asyncio.sleep(0.1)
    await b.shutdown()


@pytest.mark.asyncio
async def test_plus_will_not_receive_dollar():
    """A subscription to “+/monitor/Clients” will not receive any messages published to “$SYS/monitor/Clients [MQTT-4.7.2-1]"""
    # BrokerSysPlugin doesn't use $SYS/monitor/Clients, so this is an equivalent test with $SYS/broker topics

    cfg = {
        'listeners': {'default': {'type': 'tcp', 'bind': '127.0.0.1'}},
        'plugins': {
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {"allow_anonymous": True},
            'amqtt.plugins.sys.broker.BrokerSysPlugin': {"sys_interval": 2}
        }
    }

    b = Broker(config=cfg)
    await b.start()
    await asyncio.sleep(0.1)
    c = MQTTClient(config={'auto_reconnect': False})
    await c.connect()
    await asyncio.sleep(0.1)
    await c.subscribe(
        [('+/broker/#', QOS_0),
         ('+/broker/time', QOS_0),
         ('+/broker/clients/#', QOS_0),
         ('+/broker/+/maximum', QOS_0)
         ]
    )
    await asyncio.sleep(0.1)

    with pytest.raises(asyncio.TimeoutError):
        # wait long enough for broker sys plugin to run
        _ = await c.deliver_message(timeout_duration=5)

    await c.disconnect()
    await asyncio.sleep(0.1)
    await b.shutdown()



import logging
from pathlib import Path
import sys

import typer

logger = logging.getLogger(__name__)

app = typer.Typer(add_completion=False, rich_markup_mode=None)


def main() -> None:
    """Run the cli for `ca_creds`."""
    app()


@app.command()
def ca_creds(
        country: str = typer.Option(..., "--country", help="x509 'country_name' attribute"),
        state: str = typer.Option(..., "--state", help="x509 'state_or_province_name' attribute"),
        locality: str = typer.Option(..., "--locality", help="x509 'locality_name' attribute"),
        org_name: str = typer.Option(..., "--org-name", help="x509 'organization_name' attribute"),
        cn: str = typer.Option(..., "--cn", help="x509 'common_name' attribute"),
        output_dir: str = typer.Option(Path.cwd().absolute(), "--output-dir", help="output directory"),
) -> None:
    """Generate a self-signed key and certificate to be used as the root CA, with a key size of 2048 and a 1-year expiration."""
    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    try:
        from amqtt.contrib.cert import generate_root_creds, write_key_and_crt  # pylint: disable=import-outside-toplevel
    except ImportError:
        msg = "Requires installation of the optional 'contrib' package: `pip install amqtt[contrib]`"
        logger.critical(msg)
        sys.exit(1)

    ca_key, ca_crt = generate_root_creds(country=country, state=state, locality=locality, org_name=org_name, cn=cn)

    write_key_and_crt(ca_key, ca_crt, "ca", Path(output_dir))


if __name__ == "__main__":
    main()



import asyncio

try:
    from asyncio import InvalidStateError, QueueFull, QueueShutDown
except ImportError:
    # Fallback for Python < 3.12
    class InvalidStateError(Exception):  # type: ignore[no-redef]
        pass

    class QueueFull(Exception):  # type: ignore[no-redef]  # noqa : N818
        pass

    class QueueShutDown(Exception):  # type: ignore[no-redef]  # noqa : N818
        pass


import collections
import itertools
import logging
from typing import Generic, TypeVar, cast

from amqtt.adapters import ReaderAdapter, WriterAdapter
from amqtt.contexts import BaseContext
from amqtt.errors import AMQTTError, MQTTError, NoDataError, ProtocolHandlerError
from amqtt.events import MQTTEvents
from amqtt.mqtt import packet_class
from amqtt.mqtt.connack import ConnackPacket
from amqtt.mqtt.connect import ConnectPacket
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.disconnect import DisconnectPacket
from amqtt.mqtt.packet import (
    CONNACK,
    CONNECT,
    DISCONNECT,
    PINGREQ,
    PINGRESP,
    PUBACK,
    PUBCOMP,
    PUBLISH,
    PUBREC,
    PUBREL,
    RESERVED_0,
    RESERVED_15,
    SUBACK,
    SUBSCRIBE,
    UNSUBACK,
    UNSUBSCRIBE,
    MQTTFixedHeader,
)
from amqtt.mqtt.pingreq import PingReqPacket
from amqtt.mqtt.pingresp import PingRespPacket
from amqtt.mqtt.puback import PubackPacket
from amqtt.mqtt.pubcomp import PubcompPacket
from amqtt.mqtt.publish import PublishPacket
from amqtt.mqtt.pubrec import PubrecPacket
from amqtt.mqtt.pubrel import PubrelPacket
from amqtt.mqtt.suback import SubackPacket
from amqtt.mqtt.subscribe import SubscribePacket
from amqtt.mqtt.unsuback import UnsubackPacket
from amqtt.mqtt.unsubscribe import UnsubscribePacket
from amqtt.plugins.manager import PluginManager
from amqtt.session import INCOMING, OUTGOING, ApplicationMessage, IncomingApplicationMessage, OutgoingApplicationMessage, Session

C = TypeVar("C", bound=BaseContext)


class ProtocolHandler(Generic[C]):
    """Class implementing the MQTT communication protocol using asyncio features."""

    def __init__(
        self,
        plugins_manager: PluginManager[C],
        session: Session | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        self.logger: logging.Logger | logging.LoggerAdapter[logging.Logger] = logging.getLogger(__name__)
        if session is not None:
            self._init_session(session)
        else:
            self.session: Session | None = None
        self.reader: ReaderAdapter | None = None
        self.writer: WriterAdapter | None = None
        self.plugins_manager: PluginManager[C] = plugins_manager

        try:
            self._loop = loop if loop is not None else asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self._reader_task: asyncio.Task[None] | None = None
        self._keepalive_task: asyncio.TimerHandle | None = None
        self._reader_ready: asyncio.Event | None = None
        self._reader_stopped = asyncio.Event()
        self._puback_waiters: dict[int, asyncio.Future[PubackPacket]] = {}
        self._pubrec_waiters: dict[int, asyncio.Future[PubrecPacket]] = {}
        self._pubrel_waiters: dict[int, asyncio.Future[PubrelPacket]] = {}
        self._pubcomp_waiters: dict[int, asyncio.Future[PubcompPacket]] = {}
        self._write_lock = asyncio.Lock()

    def _init_session(self, session: Session) -> None:
        if not session:
            msg = "Session cannot be None"
            raise AMQTTError(msg)
        log = logging.getLogger(__name__)
        self.session = session
        self.logger = logging.LoggerAdapter(log, {"client_id": self.session.client_id})
        self.keepalive_timeout: int | None = self.session.keep_alive
        if self.keepalive_timeout <= 0:
            self.keepalive_timeout = None

    def attach(self, session: Session, reader: ReaderAdapter, writer: WriterAdapter) -> None:
        if self.session:
            msg = "Handler is already attached to a session"
            raise ProtocolHandlerError(msg)
        self._init_session(session)
        self.reader = reader
        self.writer = writer

    def detach(self) -> None:
        self.session = None
        self.reader = None
        self.writer = None

    def _is_attached(self) -> bool:
        return bool(self.session)

    async def start(self) -> None:
        if not self._is_attached():
            msg = "Handler is not attached to a stream"
            raise ProtocolHandlerError(msg)
        self._reader_ready = asyncio.Event()
        self._reader_stopped = asyncio.Event()
        self._reader_task = asyncio.create_task(self._reader_loop())
        await self._reader_ready.wait()
        if self._loop is not None and self.keepalive_timeout is not None:
            self._keepalive_task = self._loop.call_later(self.keepalive_timeout, self.handle_write_timeout)
        self.logger.debug("Handler tasks started")
        await self._retry_deliveries()
        self.logger.debug("Handler ready")

    async def stop(self) -> None:
        # Stop messages flow waiter
        self._stop_waiters()
        if self._keepalive_task:
            self._keepalive_task.cancel()
        self.logger.debug("Waiting for tasks to be stopped")
        if self._reader_task and not self._reader_task.done():
            self._reader_task.cancel()
            await self._reader_stopped.wait()
        self.logger.debug("Closing writer")
        try:
            if self.writer is not None:
                await self.writer.close()
        except asyncio.CancelledError:
            # canceling the task is the expected result
            self.logger.debug("Writer close was cancelled.")
        except asyncio.TimeoutError:
            self.logger.debug("Writer close operation timed out.", exc_info=True)
        except OSError:
            self.logger.debug("Writer close failed due to I/O error.", exc_info=True)

    def _stop_waiters(self) -> None:
        self.logger.debug(f"Stopping {len(self._puback_waiters)} puback waiters")
        self.logger.debug(f"Stopping {len(self._pubcomp_waiters)} pucomp waiters")
        self.logger.debug(f"Stopping {len(self._pubrec_waiters)} purec waiters")
        self.logger.debug(f"Stopping {len(self._pubrel_waiters)} purel waiters")
        for waiter in itertools.chain(
            self._puback_waiters.values(),
            self._pubcomp_waiters.values(),
            self._pubrec_waiters.values(),
            self._pubrel_waiters.values(),
        ):
            if not isinstance(waiter, asyncio.Future):
                msg = "Waiter is not a asyncio.Future"
                raise AMQTTError(msg)
            waiter.cancel()

    async def _retry_deliveries(self) -> None:
        """Handle [MQTT-4.4.0-1] by resending PUBLISH and PUBREL messages for pending out messages."""
        self.logger.debug("Begin messages delivery retries")
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        tasks = [
            asyncio.create_task(
                asyncio.wait_for(
                    self._handle_message_flow(cast("IncomingApplicationMessage | OutgoingApplicationMessage", message)),
                    10,
                ),
            )
            for message in itertools.chain(self.session.inflight_in.values(), self.session.inflight_out.values())
        ]
        if tasks:
            done, pending = await asyncio.wait(tasks)
            self.logger.debug(f"{len(done)} messages redelivered")
            self.logger.debug(f"{len(pending)} messages not redelivered due to timeout")
        self.logger.debug("End messages delivery retries")

    async def mqtt_publish(
        self,
        topic: str,
        data: bytes | bytearray,
        qos: int | None,
        retain: bool,
        ack_timeout: int | None = None,
    ) -> OutgoingApplicationMessage:
        """Send a MQTT publish message and manage messages flows.

        This method doesn't return until the message has been acknowledged by receiver or timeout occurs.
        :param topic: MQTT topic to publish
        :param data:  data to send on topic
        :param qos: quality of service to use for message flow. Can be QOS_0, QOS_1 or QOS_2
        :param retain: retain message flag
        :param ack_timeout: acknowledge timeout. If set, this method will return a TimeOut error if the acknowledgment
        is not completed before ack_timeout second
        :return: ApplicationMessage used during inflight operations.
        """
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        if qos in (QOS_1, QOS_2):
            packet_id = self.session.next_packet_id
            if packet_id in self.session.inflight_out:
                msg = f"A message with the same packet ID '{packet_id}' is already in flight"
                raise AMQTTError(msg)
        else:
            packet_id = None
        message: OutgoingApplicationMessage = OutgoingApplicationMessage(packet_id, topic, qos, data, retain)
        # Handle message flow
        if ack_timeout is not None and ack_timeout > 0:
            await asyncio.wait_for(self._handle_message_flow(message), ack_timeout)
        else:
            await self._handle_message_flow(message)
        return message

    async def _handle_message_flow(self, app_message: IncomingApplicationMessage | OutgoingApplicationMessage) -> None:
        """Handle protocol flow for incoming and outgoing messages.

        Depending on service level and according to MQTT spec. paragraph 4.3-Quality of Service levels and protocol flows.
        :param app_message: PublishMessage to handle
        """
        if app_message.qos not in (QOS_0, QOS_1, QOS_2):
            msg = f"Unexpected QOS value '{app_message.qos}' for message: {app_message}"
            raise AMQTTError(msg)
        if app_message.qos == QOS_0:
            await self._handle_qos0_message_flow(app_message)
        elif app_message.qos == QOS_1:
            await self._handle_qos1_message_flow(app_message)
        elif app_message.qos == QOS_2:
            await self._handle_qos2_message_flow(app_message)
        else:
            msg = f"Unexpected QOS value '{app_message.qos}'"
            raise AMQTTError(msg)

    async def _handle_qos0_message_flow(self, app_message: IncomingApplicationMessage | OutgoingApplicationMessage) -> None:
        """Handle QOS_0 application message acknowledgment.

        For incoming messages, this method stores the message.
        For outgoing messages, this methods sends PUBLISH.
        :param app_message: Application message to handle
        """
        if app_message.qos != QOS_0:
            msg = f"Expected QOS_0 message, got QOS_{app_message.qos}"
            raise ValueError(msg)
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        if app_message.direction == OUTGOING:
            packet = app_message.build_publish_packet()
            # Send PUBLISH packet
            await self._send_packet(packet)
            app_message.publish_packet = packet
        elif app_message.direction == INCOMING:
            if app_message.publish_packet is not None and app_message.publish_packet.dup_flag:
                self.logger.warning(
                    "[MQTT-3.3.1-2] DUP flag must set to 0 for QOS 0 message. Message ignored: %r",
                    app_message.publish_packet,
                )
            else:
                try:
                    self.session.delivered_message_queue.put_nowait(app_message)
                    self.logger.debug(f"Message added to delivery queue: {app_message}")
                except QueueShutDown as e:
                    self.logger.warning(f"Delivered messages queue is shut down. QOS_0 message discarded: {e}")
                except QueueFull as e:
                    self.logger.warning(f"Delivered messages queue is full. QOS_0 message discarded: {e}")

    async def _handle_qos1_message_flow(self, app_message: OutgoingApplicationMessage | IncomingApplicationMessage) -> None:
        """Handle QOS_1 application message acknowledgment.

        For incoming messages, this method stores the message and reply with PUBACK.
        For outgoing messages, this methods sends PUBLISH and waits for the corresponding PUBACK.
        :param app_message: Application message to handle
        """
        if app_message.qos != QOS_1:
            msg = f"Expected QOS_1 message, got QOS_{app_message.qos}"
            raise ValueError(msg)
        if app_message.packet_id is None:
            msg = "Packet ID is not set"
            raise ValueError(msg)
        if app_message.puback_packet:
            msg = f"Message '{app_message.packet_id}' has already been acknowledged"
            raise AMQTTError(msg)
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)

        if app_message.direction == OUTGOING:
            if app_message.packet_id not in self.session.inflight_out and isinstance(app_message, OutgoingApplicationMessage):
                # Store message in session
                self.session.inflight_out[app_message.packet_id] = app_message
            if app_message.publish_packet is not None:
                # A Publish packet has already been sent, this is a retry
                publish_packet = app_message.build_publish_packet(dup=True)
            else:
                publish_packet = app_message.build_publish_packet()
            # Send PUBLISH packet
            await self._send_packet(publish_packet)
            app_message.publish_packet = publish_packet
            # Wait for puback
            waiter: asyncio.Future[PubackPacket] = asyncio.Future()
            self._puback_waiters[app_message.packet_id] = waiter
            try:
                app_message.puback_packet = await asyncio.wait_for(waiter, timeout=5)
            except asyncio.TimeoutError:
                msg = f"Timeout waiting for PUBACK for packet ID {app_message.packet_id}"
                self.logger.warning(msg)
                raise TimeoutError(msg) from None
            finally:
                self._puback_waiters.pop(app_message.packet_id, None)
                # Discard inflight message
                self.session.inflight_out.pop(app_message.packet_id, None)
        elif app_message.direction == INCOMING:
            # Initiate delivery
            self.logger.debug("Add message to delivery")
            await self.session.delivered_message_queue.put(app_message)
            # Send PUBACK
            puback = PubackPacket.build(app_message.packet_id)
            await self._send_packet(puback)
            app_message.puback_packet = puback

    async def _handle_qos2_message_flow(self, app_message: OutgoingApplicationMessage | IncomingApplicationMessage) -> None:
        """Handle QOS_2 application message acknowledgment.

        For incoming messages, this method stores the message, sends PUBREC, waits for PUBREL, initiate delivery
        and send PUBCOMP.
        For outgoing messages, this methods sends PUBLISH, waits for PUBREC, discards messages and wait for PUBCOMP.
        :param app_message: Application message to handle
        """
        if app_message.qos != QOS_2:
            msg = f"Expected QOS_2 message, got QOS_{app_message.qos}"
            raise ValueError(msg)
        if app_message.packet_id is None:
            msg = "Packet ID is not set"
            raise ValueError(msg)
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)

        if app_message.direction == OUTGOING:
            if app_message.pubrel_packet and app_message.pubcomp_packet:
                msg = f"Message '{app_message.packet_id}' has already been acknowledged"
                raise AMQTTError(msg)

            if not app_message.pubrel_packet:
                # Store message
                publish_packet: PublishPacket
                if app_message.publish_packet is not None:
                    # This is a retry flow, no need to store just check the message exists in session
                    if app_message.packet_id not in self.session.inflight_out:
                        msg = f"Unknown inflight message '{app_message.packet_id}' in session"
                        raise AMQTTError(msg)
                    publish_packet = app_message.build_publish_packet(dup=True)
                elif isinstance(app_message, OutgoingApplicationMessage):
                    # Store message in session
                    self.session.inflight_out[app_message.packet_id] = app_message
                    publish_packet = app_message.build_publish_packet()
                else:
                    self.logger.debug("Message can not be stored, to be checked!")
                # Send PUBLISH packet
                await self._send_packet(publish_packet)
                app_message.publish_packet = publish_packet
                # Wait PUBREC
                if app_message.packet_id in self._pubrec_waiters:
                    # PUBREC waiter already exists for this packet ID
                    message = f"Can't add PUBREC waiter, a waiter already exists for message Id '{app_message.packet_id}'"
                    self.logger.warning(message)
                    raise AMQTTError(message)
                waiter_pub_rec: asyncio.Future[PubrecPacket] = asyncio.Future()
                self._pubrec_waiters[app_message.packet_id] = waiter_pub_rec
                try:
                    app_message.pubrec_packet = await waiter_pub_rec
                finally:
                    self._pubrec_waiters.pop(app_message.packet_id, None)
                    self.session.inflight_out.pop(app_message.packet_id, None)

            if not app_message.pubcomp_packet:
                # Send pubrel
                app_message.pubrel_packet = PubrelPacket.build(app_message.packet_id)
                await self._send_packet(app_message.pubrel_packet)
                # Wait for PUBCOMP
                waiter_pub_comp: asyncio.Future[PubcompPacket] = asyncio.Future()
                self._pubcomp_waiters[app_message.packet_id] = waiter_pub_comp
                try:
                    app_message.pubcomp_packet = await waiter_pub_comp
                finally:
                    self._pubcomp_waiters.pop(app_message.packet_id, None)
                    self.session.inflight_out.pop(app_message.packet_id, None)
        elif app_message.direction == INCOMING and isinstance(app_message, IncomingApplicationMessage):
            self.session.inflight_in[app_message.packet_id] = app_message
            # Send pubrec
            pubrec_packet = PubrecPacket.build(app_message.packet_id)
            await self._send_packet(pubrec_packet)
            app_message.pubrec_packet = pubrec_packet
            # Wait PUBREL
            if app_message.packet_id in self._pubrel_waiters and not self._pubrel_waiters[app_message.packet_id].done():
                # PUBREL waiter already exists for this packet ID
                message = f"A waiter already exists for message Id '{app_message.packet_id}', canceling it"
                self.logger.warning(message)
                self._pubrel_waiters[app_message.packet_id].cancel()
            try:
                waiter_pub_rel: asyncio.Future[PubrelPacket] = asyncio.Future()
                self._pubrel_waiters[app_message.packet_id] = waiter_pub_rel
                await waiter_pub_rel
                del self._pubrel_waiters[app_message.packet_id]
                app_message.pubrel_packet = waiter_pub_rel.result()
                # Initiate delivery and discard message
                await self.session.delivered_message_queue.put(app_message)
                del self.session.inflight_in[app_message.packet_id]
                # Send pubcomp
                pubcomp_packet = PubcompPacket.build(app_message.packet_id)
                await self._send_packet(pubcomp_packet)
                app_message.pubcomp_packet = pubcomp_packet
            except asyncio.CancelledError:
                self.logger.debug("Message flow cancelled")
        else:
            self.logger.debug("Unknown direction!")

    async def _reader_loop(self) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        if not self._reader_ready:
            msg = "Reader ready is not initialized."
            raise ProtocolHandlerError(msg)

        self.logger.debug(f"{self.session.client_id} Starting reader coro")
        running_tasks: collections.deque[asyncio.Task[None]] = collections.deque()
        keepalive_timeout: int | None = self.session.keep_alive
        if keepalive_timeout is not None and keepalive_timeout <= 0:
            keepalive_timeout = None
        while True:
            try:
                self._reader_ready.set()
                while running_tasks and running_tasks[0].done():
                    running_tasks.popleft()
                if len(running_tasks) > 1:
                    self.logger.debug(f"Handler running tasks: {len(running_tasks)}")
                if self.reader is None:
                    self.logger.warning("Reader is not initialized!")
                    break
                fixed_header = await asyncio.wait_for(MQTTFixedHeader.from_stream(self.reader), timeout=keepalive_timeout)
                if not fixed_header:
                    self.logger.debug(f"{self.session.client_id} No more data (EOF received), stopping reader coro")
                    break
                if fixed_header.packet_type in (RESERVED_0, RESERVED_15):
                    self.logger.warning(
                        f"{self.session.client_id} Received reserved packet, which is forbidden: closing connection",
                    )
                    await self.handle_connection_closed()
                    continue

                cls = packet_class(fixed_header)
                packet = await cls.from_stream(self.reader, fixed_header=fixed_header)
                await self.plugins_manager.fire_event(MQTTEvents.PACKET_RECEIVED, packet=packet, session=self.session)
                if packet.fixed_header is None or packet.fixed_header.packet_type not in (
                    CONNACK,
                    SUBSCRIBE,
                    UNSUBSCRIBE,
                    SUBACK,
                    UNSUBACK,
                    PUBACK,
                    PUBREC,
                    PUBREL,
                    PUBCOMP,
                    PINGREQ,
                    PINGRESP,
                    PUBLISH,
                    DISCONNECT,
                    CONNECT,
                ):
                    self.logger.warning(f"{self.session.client_id} Unhandled packet type: {packet.fixed_header.packet_type}")
                    continue

                task: asyncio.Task[None] | None = None
                if packet.fixed_header.packet_type == CONNACK and isinstance(packet, ConnackPacket):
                    task = asyncio.create_task(self.handle_connack(packet))
                elif packet.fixed_header.packet_type == SUBSCRIBE and isinstance(packet, SubscribePacket):
                    task = asyncio.create_task(self.handle_subscribe(packet))
                elif packet.fixed_header.packet_type == UNSUBSCRIBE and isinstance(packet, UnsubscribePacket):
                    task = asyncio.create_task(self.handle_unsubscribe(packet))
                elif packet.fixed_header.packet_type == SUBACK and isinstance(packet, SubackPacket):
                    task = asyncio.create_task(self.handle_suback(packet))
                elif packet.fixed_header.packet_type == UNSUBACK and isinstance(packet, UnsubackPacket):
                    task = asyncio.create_task(self.handle_unsuback(packet))
                elif packet.fixed_header.packet_type == PUBACK and isinstance(packet, PubackPacket):
                    task = asyncio.create_task(self.handle_puback(packet))
                elif packet.fixed_header.packet_type == PUBREC and isinstance(packet, PubrecPacket):
                    task = asyncio.create_task(self.handle_pubrec(packet))
                elif packet.fixed_header.packet_type == PUBREL and isinstance(packet, PubrelPacket):
                    task = asyncio.create_task(self.handle_pubrel(packet))
                elif packet.fixed_header.packet_type == PUBCOMP and isinstance(packet, PubcompPacket):
                    task = asyncio.create_task(self.handle_pubcomp(packet))
                elif packet.fixed_header.packet_type == PINGREQ and isinstance(packet, PingReqPacket):
                    task = asyncio.create_task(self.handle_pingreq(packet))
                elif packet.fixed_header.packet_type == PINGRESP and isinstance(packet, PingRespPacket):
                    task = asyncio.create_task(self.handle_pingresp(packet))
                elif packet.fixed_header.packet_type == PUBLISH and isinstance(packet, PublishPacket):
                    task = asyncio.create_task(self.handle_publish(packet))
                elif packet.fixed_header.packet_type == DISCONNECT and isinstance(packet, DisconnectPacket):
                    task = asyncio.create_task(self.handle_disconnect(packet))
                elif packet.fixed_header.packet_type == CONNECT and isinstance(packet, ConnectPacket):
                    # q: why is this not like all other inside a create_task?
                    # a: the connection needs to be established before any other packet tasks for this new session are scheduled
                    await self.handle_connect(packet)
                if task:
                    running_tasks.append(task)
            except MQTTError:
                self.logger.debug("Message discarded")
            except asyncio.CancelledError:
                self.logger.debug("Task cancelled, reader loop ending")
                break
            except asyncio.TimeoutError:
                self.logger.debug(f"{self.session.client_id} Input stream read timeout")
                self.handle_read_timeout()
            except NoDataError:
                self.logger.debug(f"{self.session.client_id} No data available")
            except Exception as e:  # noqa: BLE001, pylint: disable=W0718
                self.logger.warning(f"{type(self).__name__} Unhandled exception in reader coro: {e!r}")
                break
        while running_tasks:
            running_tasks.popleft().cancel()
        await self.handle_connection_closed()
        self._reader_stopped.set()
        self.logger.debug("Reader coro stopped")
        await self.stop()

    async def _send_packet(
        self,
        packet: PublishPacket
        | PubackPacket
        | ConnackPacket
        | SubackPacket
        | ConnectPacket
        | SubscribePacket
        | UnsubscribePacket
        | DisconnectPacket
        | PingReqPacket
        | PubrelPacket
        | PubrecPacket
        | PubcompPacket
        | PingRespPacket
        | UnsubackPacket,
    ) -> None:
        try:
            if self.writer:
                async with self._write_lock:
                    await packet.to_stream(self.writer)
            if self._keepalive_task:
                self._keepalive_task.cancel()
                if self.keepalive_timeout is not None:
                    self._keepalive_task = self._loop.call_later(self.keepalive_timeout, self.handle_write_timeout)
            await self.plugins_manager.fire_event(MQTTEvents.PACKET_SENT, packet=packet, session=self.session)
        except (ConnectionResetError, BrokenPipeError):
            await self.handle_connection_closed()
        except asyncio.CancelledError as e:
            msg = "Packet handling was cancelled"
            raise ProtocolHandlerError(msg) from e
        except Exception as e:
            self.logger.warning(f"Unhandled exception: {e}")
            raise

    async def mqtt_deliver_next_message(self) -> ApplicationMessage | None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)

        if not self._is_attached():
            return None
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"{self.session.delivered_message_queue.qsize()} message(s) available for delivery")
        message: ApplicationMessage | None = None
        try:
            message = await self.session.delivered_message_queue.get()
        except (asyncio.CancelledError, RuntimeError):
            message = None
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"Delivering message {message}")
        return message

    def handle_write_timeout(self) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} write timeout unhandled")

    def handle_read_timeout(self) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} read timeout unhandled")

    async def handle_connack(self, connack: ConnackPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} CONNACK unhandled")

    async def handle_connect(self, connect: ConnectPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} CONNECT unhandled")

    async def handle_subscribe(self, subscribe: SubscribePacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} SUBSCRIBE unhandled")

    async def handle_unsubscribe(self, unsubscribe: UnsubscribePacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} UNSUBSCRIBE unhandled")

    async def handle_suback(self, suback: SubackPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} SUBACK unhandled")

    async def handle_unsuback(self, unsuback: UnsubackPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} UNSUBACK unhandled")

    async def handle_pingresp(self, pingresp: PingRespPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} PINGRESP unhandled")

    async def handle_pingreq(self, pingreq: PingReqPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} PINGREQ unhandled")

    async def handle_disconnect(self, disconnect: DisconnectPacket) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} DISCONNECT unhandled")

    async def handle_connection_closed(self) -> None:
        if self.session is None:
            msg = "Session is not initialized."
            raise AMQTTError(msg)
        self.logger.debug(f"{self.session.client_id} Connection closed unhandled")

    async def handle_puback(self, puback: PubackPacket) -> None:
        if puback.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        packet_id = puback.variable_header.packet_id
        try:
            waiter = self._puback_waiters[packet_id]
            waiter.set_result(puback)
        except KeyError:
            self.logger.warning(f"Received PUBACK for unknown pending message Id: '{packet_id}'")
        except InvalidStateError:
            self.logger.warning(f"PUBACK waiter with Id '{packet_id}' already done")

    async def handle_pubrec(self, pubrec: PubrecPacket) -> None:
        packet_id = pubrec.packet_id
        try:
            waiter = self._pubrec_waiters[packet_id]
            waiter.set_result(pubrec)
        except KeyError:
            self.logger.warning(f"Received PUBREC for unknown pending message with Id: {packet_id}")
        except InvalidStateError:
            self.logger.warning(f"PUBREC waiter with Id '{packet_id}' already done")

    async def handle_pubcomp(self, pubcomp: PubcompPacket) -> None:
        packet_id = pubcomp.packet_id
        try:
            waiter = self._pubcomp_waiters[packet_id]
            waiter.set_result(pubcomp)
        except KeyError:
            self.logger.warning(f"Received PUBCOMP for unknown pending message with Id: {packet_id}")
        except InvalidStateError:
            self.logger.warning(f"PUBCOMP waiter with Id '{packet_id}' already done")

    async def handle_pubrel(self, pubrel: PubrelPacket) -> None:
        packet_id = pubrel.packet_id
        try:
            waiter = self._pubrel_waiters[packet_id]
            waiter.set_result(pubrel)
        except KeyError:
            self.logger.warning(f"Received PUBREL for unknown pending message with Id: {packet_id}")
        except InvalidStateError:
            self.logger.warning(f"PUBREL waiter with Id '{packet_id}' already done")

    async def handle_publish(self, publish_packet: PublishPacket) -> None:
        packet_id = publish_packet.variable_header.packet_id if publish_packet.variable_header else None
        qos = publish_packet.qos
        if publish_packet.topic_name is None or publish_packet.data is None:
            return
        incoming_message = IncomingApplicationMessage(
            packet_id,
            publish_packet.topic_name,
            qos,
            publish_packet.data,
            publish_packet.retain_flag,
        )
        incoming_message.publish_packet = publish_packet
        await self._handle_message_flow(incoming_message)
        self.logger.debug(f"Message queue size: {self.session.delivered_message_queue.qsize() if self.session else None}")



from dataclasses import dataclass, field, fields, replace
import logging
import warnings

try:
    from enum import Enum, StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  # type: ignore[no-redef]
        pass

from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

from dacite import Config as DaciteConfig, from_dict as dict_to_dataclass

from amqtt.mqtt.constants import QOS_0, QOS_2

if TYPE_CHECKING:
    import asyncio

logger = logging.getLogger(__name__)


class BaseContext:
    def __init__(self) -> None:
        self.loop: asyncio.AbstractEventLoop | None = None
        self.logger: logging.Logger = logging.getLogger(__name__)
        # cleanup with a `Generic` type
        self.config: ClientConfig | BrokerConfig | dict[str, Any] | None = None


class Action(StrEnum):
    """Actions issued by the broker."""

    SUBSCRIBE = "subscribe"
    PUBLISH = "publish"
    RECEIVE = "receive"


class ListenerType(StrEnum):
    """Types of mqtt listeners."""

    TCP = "tcp"
    WS = "ws"
    EXTERNAL = "external"

    def __repr__(self) -> str:
        """Display the string value, instead of the enum member."""
        return f'"{self.value!s}"'


class Dictable:
    """Add dictionary methods to a dataclass."""

    def __getitem__(self, key: str) -> Any:
        """Allow dict-style `[]` access to a dataclass."""
        return self.get(key)

    def get(self, name: str, default: Any = None) -> Any:
        """Allow dict-style access to a dataclass."""
        name = name.replace("-", "_")
        if hasattr(self, name):
            return getattr(self, name)
        if default is not None:
            return default
        msg = f"'{name}' is not defined"
        raise ValueError(msg)

    def __contains__(self, name: str) -> bool:
        """Provide dict-style 'in' check."""
        return getattr(self, name.replace("-", "_"), None) is not None

    def __iter__(self) -> Iterator[Any]:
        """Provide dict-style iteration."""
        for f in fields(self):   # type: ignore[arg-type]
            yield getattr(self, f.name)

    def copy(self) -> dataclass:  # type: ignore[valid-type]
        """Return a copy of the dataclass."""
        return replace(self)  # type: ignore[type-var]

    @staticmethod
    def _coerce_lists(value: list[Any] | dict[str, Any] | Any) -> list[dict[str, Any]]:
        if isinstance(value, list):
            return value  # It's already a list of dicts
        if isinstance(value, dict):
            return [value]  # Promote single dict to a list
        msg = "Could not convert 'list' to 'list[dict[str, Any]]'"
        raise ValueError(msg)


@dataclass
class ListenerConfig(Dictable):
    """Structured configuration for a broker's listeners."""

    type: ListenerType = ListenerType.TCP
    """Type of listener: `tcp` for 'mqtt' or `ws` for 'websocket' when specified in dictionary or yaml.'"""
    bind: str | None = "0.0.0.0:1883"
    """address and port for the listener to bind to"""
    max_connections: int = 0
    """max number of connections allowed for this listener"""
    ssl: bool = False
    """secured by ssl"""
    cafile: str | Path | None = None
    """Path to a file of concatenated CA certificates in PEM format. See
    [Certificates](https://docs.python.org/3/library/ssl.html#ssl-certificates) for more info."""
    capath: str | Path | None = None
    """Path to a directory containing one or more CA certificates in PEM format, following the
     [OpenSSL-specific layout](https://docs.openssl.org/master/man3/SSL_CTX_load_verify_locations/)."""
    cadata: str | Path | None = None
    """Either an ASCII string of one or more PEM-encoded certificates or a bytes-like object of DER-encoded certificates."""
    certfile: str | Path | None = None
    """Full path to file in PEM format containing the server's certificate (as well as any number of CA
    certificates needed to establish the certificate's authenticity.)"""
    keyfile: str | Path | None = None
    """Full path to file in PEM format containing the server's private key."""
    reader: str | None = None
    writer: str | None = None

    def __post_init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if (self.certfile is None) ^ (self.keyfile is None):
            msg = "If specifying the 'certfile' or 'keyfile', both are required."
            raise ValueError(msg)

        for fn in ("cafile", "capath", "certfile", "keyfile"):
            if isinstance(getattr(self, fn), str):
                setattr(self, fn, Path(getattr(self, fn)))
            if getattr(self, fn) and not getattr(self, fn).exists():
                msg = f"'{fn}' does not exist : {getattr(self, fn)}"
                raise FileNotFoundError(msg)

    def apply(self, other: "ListenerConfig") -> None:
        """Apply the field from 'other', if 'self' field is default."""
        for f in fields(self):
            if getattr(self, f.name) == f.default:
                setattr(self, f.name, other[f.name])


def default_listeners() -> dict[str, Any]:
    """Create defaults for BrokerConfig.listeners."""
    return {
        "default": ListenerConfig()
    }


def default_broker_plugins() -> dict[str, Any]:
    """Create defaults for BrokerConfig.plugins."""
    return {
        "amqtt.plugins.logging_amqtt.EventLoggerPlugin": {},
        "amqtt.plugins.logging_amqtt.PacketLoggerPlugin": {},
        "amqtt.plugins.authentication.AnonymousAuthPlugin": {"allow_anonymous": True},
        "amqtt.plugins.sys.broker.BrokerSysPlugin": {"sys_interval": 20}
    }


@dataclass
class BrokerConfig(Dictable):
    """Structured configuration for a broker. Can be passed directly to `amqtt.broker.Broker` or created from a dictionary."""

    listeners: dict[Literal["default"] | str, ListenerConfig] = field(default_factory=default_listeners)  # noqa: PYI051
    """Network of listeners used by the services. a 'default' named listener is required; if another listener
     does not set a value, the 'default' settings are applied. See
     [`ListenerConfig`](broker_config.md#amqtt.contexts.ListenerConfig) for more information."""
    sys_interval: int | None = None
    """*Deprecated field to configure the `BrokerSysPlugin`. See [`BrokerSysPlugin`](../plugins/packaged_plugins.md#sys-topics)
    for recommended configuration.*"""
    timeout_disconnect_delay: int | None = 0
    """Client disconnect timeout without a keep-alive."""
    session_expiry_interval: int | None = None
    """Seconds for an inactive session to be retained."""
    auth: dict[str, Any] | None = None
    """*Deprecated field used to config EntryPoint-loaded plugins. See
    [`AnonymousAuthPlugin`](../plugins/packaged_plugins.md#anonymous-auth-plugin) and
    [`FileAuthPlugin`](../plugins/packaged_plugins.md#password-file-auth-plugin) for recommended configuration.*"""
    topic_check: dict[str, Any] | None = None
    """*Deprecated field used to config EntryPoint-loaded plugins. See
    [`TopicTabooPlugin`](../plugins/packaged_plugins.md#taboo-topic-plugin) and
    [`TopicACLPlugin`](../plugins/packaged_plugins.md#acl-topic-plugin) for recommended configuration method.*"""
    plugins: dict[str, Any] | list[str | dict[str, Any]] | None = field(default_factory=default_broker_plugins)
    """The dictionary has a key of the dotted-module path of a class derived from `BasePlugin`, `BaseAuthPlugin`
     or `BaseTopicPlugin`; the value is a dictionary of configuration options for that plugin. See
     [custom plugins](../plugins/custom_plugins.md) for more information. `list[str | dict[str,Any]]` is deprecated but available
      to support legacy use cases."""

    def __post_init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if self.sys_interval is not None:
            logger.warning("sys_interval is deprecated, use 'plugins' to define configuration")

        if self.auth is not None or self.topic_check is not None:
            logger.warning("'auth' and 'topic-check' are deprecated, use 'plugins' to define configuration")

        default_listener = self.listeners["default"]
        for listener_name, listener in self.listeners.items():
            if listener_name == "default":
                continue
            listener.apply(default_listener)

        if isinstance(self.plugins, list):
            _plugins: dict[str, Any] = {}
            for plugin in self.plugins:
                # in case a plugin in a yaml file is listed without config map
                if isinstance(plugin, str):
                    _plugins |= {plugin: {}}
                    continue
                _plugins |= plugin
            self.plugins = _plugins

    @classmethod
    def from_dict(cls, d: dict[str, Any] | None) -> "BrokerConfig":
        """Create a broker config from a dictionary."""
        if d is None:
            return BrokerConfig()

        # patch the incoming dictionary so it can be loaded correctly
        if "topic-check" in d:
            d["topic_check"] = d["topic-check"]
            del d["topic-check"]

        # identify EntryPoint plugin loading and prevent 'plugins' from getting defaults
        if ("auth" in d or "topic-check" in d) and "plugins" not in d:
            d["plugins"] = None

        return dict_to_dataclass(data_class=BrokerConfig,
                                 data=d,
                                 config=DaciteConfig(
                                     cast=[StrEnum, ListenerType],
                                     strict=True,
                                     type_hooks={list[dict[str, Any]]: cls._coerce_lists}
                                 ))


@dataclass
class ConnectionConfig(Dictable):
    """Properties for connecting to the broker."""

    uri: str | None = "mqtt://127.0.0.1:1883"
    """URI of the broker"""
    cafile: str | Path | None = None
    """Path to a file of concatenated CA certificates in PEM format to verify broker's authenticity. See
    [Certificates](https://docs.python.org/3/library/ssl.html#ssl-certificates) for more info."""
    capath: str | Path | None = None
    """Path to a directory containing one or more CA certificates in PEM format, following the
     [OpenSSL-specific layout](https://docs.openssl.org/master/man3/SSL_CTX_load_verify_locations/)."""
    cadata: str | None = None
    """The certificate to verify the broker's authenticity in an ASCII string format of one or more PEM-encoded
    certificates or a bytes-like object of DER-encoded certificates."""
    certfile: str | Path | None = None
    """Full path to file in PEM format containing the client's certificate (as well as any number of CA
        certificates needed to establish the certificate's authenticity.)"""
    keyfile: str | Path | None = None
    """Full path to file in PEM format containing the client's private key associated with the certfile."""

    def __post__init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if (self.certfile is None) ^ (self.keyfile is None):
            msg = "If specifying the 'certfile' or 'keyfile', both are required."
            raise ValueError(msg)

        for fn in ("cafile", "capath", "certfile", "keyfile"):
            if isinstance(getattr(self, fn), str):
                setattr(self, fn, Path(getattr(self, fn)))


@dataclass
class TopicConfig(Dictable):
    """Configuration of how messages to specific topics are published.

    The topic name is specified as the key in the dictionary of the `ClientConfig.topics.
    """

    qos: int = 0
    """The quality of service associated with the publishing to this topic."""
    retain: bool = False
    """Determines if the message should be retained by the topic it was published."""

    def __post__init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if self.qos is not None and (self.qos < QOS_0 or self.qos > QOS_2):
            msg = "Topic config: default QoS must be 0, 1 or 2."
            raise ValueError(msg)


@dataclass
class WillConfig(Dictable):
    """Configuration of the 'last will & testament' of the client upon improper disconnection."""

    topic: str
    """The will message will be published to this topic."""
    message: str
    """The contents of the message to be published."""
    qos: int | None = QOS_0
    """The quality of service associated with sending this message."""
    retain: bool | None = False
    """Determines if the message should be retained by the topic it was published."""

    def __post__init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if self.qos is not None and (self.qos < QOS_0 or self.qos > QOS_2):
            msg = "Will config: default QoS must be 0, 1 or 2."
            raise ValueError(msg)


def default_client_plugins() -> dict[str, Any]:
    """Create defaults for `ClientConfig.plugins`."""
    return {
        "amqtt.plugins.logging_amqtt.PacketLoggerPlugin": {}
    }


@dataclass
class ClientConfig(Dictable):
    """Structured configuration for a broker. Can be passed directly to `amqtt.broker.Broker` or created from a dictionary."""

    keep_alive: int | None = 10
    """Keep-alive timeout sent to the broker."""
    ping_delay: int | None = 1
    """Auto-ping delay before keep-alive timeout. Setting to 0 will disable which may lead to broker disconnection."""
    default_qos: int | None = QOS_0
    """Default QoS for messages published."""
    default_retain: bool | None = False
    """Default retain value to messages published."""
    auto_reconnect: bool | None = True
    """Enable or disable auto-reconnect if connection with the broker is interrupted."""
    connection_timeout: int | None = 60
    """The number of seconds before a connection times out"""
    reconnect_retries: int | None = 2
    """Number of reconnection retry attempts. Negative value will cause client to reconnect indefinitely."""
    reconnect_max_interval: int | None = 10
    """Maximum seconds to wait before retrying a connection."""
    cleansession: bool | None = True
    """Upon reconnect, should subscriptions be cleared. Can be overridden by `MQTTClient.connect`"""
    topics: dict[str, TopicConfig] | None = field(default_factory=dict)
    """Specify the topics and what flags should be set for messages published to them."""
    broker: ConnectionConfig | None = None
    """*Deprecated* Configuration for connecting to the broker. Use `connection` field instead."""
    connection: ConnectionConfig = field(default_factory=ConnectionConfig)
    """Configuration for connecting to the broker. See
     [`ConnectionConfig`](client_config.md#amqtt.contexts.ConnectionConfig) for more information."""
    plugins: dict[str, Any] | list[dict[str, Any]] | None = field(default_factory=default_client_plugins)
    """The dictionary has a key of the dotted-module path of a class derived from `BasePlugin`; the value is
     a dictionary of configuration options for that plugin. See [custom plugins](../plugins/custom_plugins.md) for
      more information. `list[str | dict[str,Any]]` is deprecated but available to support legacy use cases."""
    check_hostname: bool | None = True
    """If establishing a secure connection, should the hostname of the certificate be verified."""
    will: WillConfig | None = None
    """Message, topic and flags that should be sent to if the client disconnects. See
    [`WillConfig`](client_config.md#amqtt.contexts.WillConfig) for more information."""

    def __post_init__(self) -> None:
        """Check config for errors and transform fields for easier use."""
        if self.default_qos is not None and (self.default_qos < QOS_0 or self.default_qos > QOS_2):
            msg = "Client config: default QoS must be 0, 1 or 2."
            raise ValueError(msg)

        if self.broker is not None:
            warnings.warn("The 'broker' option is deprecated, please use 'connection' instead.", stacklevel=2)
            self.connection = self.broker

        if bool(not self.connection.keyfile) ^ bool(not self.connection.certfile):
            msg = "Connection key and certificate files are _both_ required."
            raise ValueError(msg)

    @classmethod
    def from_dict(cls, d: dict[str, Any] | None) -> "ClientConfig":
        """Create a client config from a dictionary."""
        if d is None:
            return ClientConfig()

        return dict_to_dataclass(data_class=ClientConfig,
                                 data=d,
                                 config=DaciteConfig(
                                     cast=[StrEnum],
                                     strict=True)
                                 )



import asyncio
from collections import deque
from collections.abc import Callable, Coroutine
import contextlib
from functools import wraps
import logging
import ssl
from typing import TYPE_CHECKING, Any, TypeAlias, cast
from urllib.parse import urlparse, urlunparse

import websockets
from websockets import HeadersLike, InvalidHandshake, InvalidURI

from amqtt.adapters import (
    StreamReaderAdapter,
    StreamWriterAdapter,
    WebSocketsReader,
    WebSocketsWriter,
)
from amqtt.contexts import BaseContext, ClientConfig
from amqtt.errors import ClientError, ConnectError, ProtocolHandlerError
from amqtt.mqtt.connack import CONNECTION_ACCEPTED
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.protocol.client_handler import ClientProtocolHandler
from amqtt.plugins.manager import PluginManager
from amqtt.session import ApplicationMessage, OutgoingApplicationMessage, Session
from amqtt.utils import gen_client_id

if TYPE_CHECKING:
    from websockets.asyncio.client import ClientConnection


class ClientContext(BaseContext):
    """ClientContext is used as the context passed to plugins interacting with the client.

    It acts as an adapter to client services from plugins.
    """

    def __init__(self) -> None:
        super().__init__()
        self.config: ClientConfig | None = None


base_logger = logging.getLogger(__name__)

_F: TypeAlias = Callable[..., Coroutine[Any, Any, Any]]


def mqtt_connected(func: _F) -> _F:
    """MQTTClient coroutines decorator which will wait until connection before calling the decorated method.

    :param func: coroutine to be called once connected
    :return: coroutine result.
    """

    @wraps(func)
    async def wrapper(self: "MQTTClient", *args: Any, **kwargs: Any) -> Any:
        if not self._connected_state.is_set():
            base_logger.warning("Client not connected, waiting for it")
            _, pending = await asyncio.wait(
                [
                    asyncio.create_task(self._connected_state.wait()),
                    asyncio.create_task(self._no_more_connections.wait()),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for t in pending:
                t.cancel()
            if self._no_more_connections.is_set():
                msg = "Will not reconnect"
                raise ClientError(msg)
        return await func(self, *args, **kwargs)

    return cast("_F", wrapper)


class MQTTClient:
    """MQTT client implementation, providing an API for connecting to a broker and send/receive messages using the MQTT protocol.

    Args:
        client_id: MQTT client ID to use when connecting to the broker. If none,
            it will be generated randomly by `amqtt.utils.gen_client_id`
        config: `ClientConfig` or dictionary of equivalent structure options (see [client configuration](client_config.md)).

    Raises:
        PluginImportError: if importing a plugin from configuration fails
        PluginInitError: if initialization plugin fails

    """

    def __init__(self, client_id: str | None = None, config: ClientConfig | dict[str, Any] | None = None) -> None:
        self.logger = logging.getLogger(__name__)

        if isinstance(config, dict):
            self.config = ClientConfig.from_dict(config)
        else:
            self.config = config or ClientConfig()

        self.client_id = client_id if client_id is not None else gen_client_id()

        self.session: Session | None = None
        self._handler: ClientProtocolHandler | None = None
        self._disconnect_task: asyncio.Task[Any] | None = None
        self._connected_state = asyncio.Event()
        self._no_more_connections = asyncio.Event()
        self.additional_headers: dict[str, Any] | HeadersLike = {}

        # Init plugins manager
        context = ClientContext()
        context.config = self.config
        self.plugins_manager: PluginManager[ClientContext] = PluginManager("amqtt.client.plugins", context)
        self.client_tasks: deque[asyncio.Task[Any]] = deque()

    async def connect(
        self,
        uri: str | None = None,
        cleansession: bool | None = None,
        cafile: str | None = None,
        capath: str | None = None,
        cadata: str | None = None,
        additional_headers: dict[str, Any] | HeadersLike | None = None,
    ) -> int:
        """Connect to a remote broker.

        At first, a network connection is established with the server
        using the given protocol (``mqtt``, ``mqtts``, ``ws`` or ``wss``).
        Once the socket is connected, a
        [CONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>)
        message is sent with the requested information.

        Args:
            uri: Broker URI connection, conforming to
                [MQTT URI scheme](https://github.com/mqtt/mqtt.github.io/wiki/URI-Scheme). default,
                will be taken from the ``uri`` config attribute.
            cleansession: MQTT CONNECT clean session flag
            cafile: server certificate authority file (optional, used for secured connection)
            capath: server certificate authority path (optional, used for secured connection)
            cadata: server certificate authority data (optional, used for secured connection)
            additional_headers: a dictionary with additional http headers that should be sent on the
                initial connection (optional, used only with websocket connections)

        Returns:
            [CONNACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033)'s return code

        Raises:
            ConnectError: could not connect to broker

        """
        additional_headers = additional_headers if additional_headers is not None else {}
        self.session = self._init_session(uri, cleansession, cafile, capath, cadata)
        self.additional_headers = additional_headers
        self.logger.debug(f"Connecting to: {self.session.broker_uri}")

        try:
            return await self._do_connect()
        except asyncio.CancelledError as e:
            msg = "Future or Task was cancelled"
            raise ConnectError(msg) from e
        # no matter the failure mode, still try to reconnect
        except Exception as e:  # pylint: disable=W0718
            self.logger.warning(f"Connection failed: {e!r}")
            if not self.config.get("auto_reconnect", False):
                raise
            return await self.reconnect()

    async def disconnect(self) -> None:
        """Disconnect from the connected broker.

        This method sends a [DISCONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718090)
        message and closes the network socket.

        """
        await self.cancel_tasks()

        if not (self.session and self._handler):
            self.logger.warning("Session or handler not initialized, ignoring disconnect.")
            return

        if not self.session.transitions.is_connected():
            self.logger.warning("Client session not connected, ignoring call.")
            return

        if self._disconnect_task and not self._disconnect_task.done():
            self._disconnect_task.cancel()

        await self._handler.mqtt_disconnect()
        self._connected_state.clear()
        await self._handler.stop()
        self.session.transitions.disconnect()

    async def cancel_tasks(self) -> None:
        """Cancel all pending tasks."""
        while self.client_tasks:
            task = self.client_tasks.pop()
            task.cancel()

    async def reconnect(self, cleansession: bool | None = None) -> int:
        """Reconnect a previously connected broker.

        Reconnection tries to establish a network connection
        and send a [CONNECT](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028) message.
        Retries interval and attempts can be controlled with the ``reconnect_max_interval``
        and ``reconnect_retries`` configuration parameters.

        Args:
            cleansession: clean session flag used in MQTT CONNECT messages sent for reconnections.

        Returns:
            [CONNACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033) return code

        Raises:
            amqtt.client.ConnectException: if re-connection fails after max retries.

        """
        if self.session and self.session.transitions.is_connected():
            self.logger.warning("Client already connected")
            return CONNECTION_ACCEPTED

        if self.session and cleansession:
            self.session.clean_session = cleansession
        self.logger.debug(f"Reconnecting with session parameters: {self.session}")

        reconnect_max_interval = self.config.get("reconnect_max_interval", 10)
        reconnect_retries = self.config.get("reconnect_retries", 2)
        nb_attempt = 1

        while True:
            try:
                self.logger.debug(f"Reconnect attempt {nb_attempt}...")
                return await self._do_connect()
            except asyncio.CancelledError as e:
                msg = "Future or Task was cancelled"
                raise ConnectError(msg) from e
            # no matter the failure mode, still try to reconnect
            except Exception as e:  # pylint: disable=W0718
                self.logger.warning(f"Reconnection attempt failed: {e!r}")
                self.logger.debug("", exc_info=True)
                if 0 <= reconnect_retries < nb_attempt:
                    self.logger.exception("Maximum connection attempts reached. Reconnection aborted.")
                    self.logger.debug("", exc_info=True)
                    msg = "Too many failed attempts"
                    raise ConnectError(msg) from e
                delay = min(reconnect_max_interval, 2**nb_attempt)
                self.logger.debug(f"Waiting {delay} seconds before next attempt")
                await asyncio.sleep(delay)
                nb_attempt += 1

    async def _do_connect(self) -> int:
        return_code = await self._connect_coro()
        self._disconnect_task = asyncio.create_task(self.handle_connection_close())
        return return_code

    @mqtt_connected
    async def ping(self) -> None:
        """Ping the broker.

        Send a MQTT [PINGREQ](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718081)
        message for response.

        """
        if self.session and self._handler and self.session.transitions.is_connected():
            await self._handler.mqtt_ping()
        elif not self.session:
            self.logger.warning("Session is not initialized.")
        elif not self._handler:
            self.logger.warning("Handler is not initialized.")
        else:
            self.logger.warning(f"PING incompatible with state '{self.session.transitions.state}'")

    @mqtt_connected
    async def publish(
        self,
        topic: str,
        message: bytes,
        qos: int | None = None,
        retain: bool | None = None,
        ack_timeout: int | None = None,
    ) -> OutgoingApplicationMessage:
        """Publish a message to the broker.

        Send a MQTT [PUBLISH](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037)
        message and wait for acknowledgment depending on Quality Of Service

        Args:
            topic: topic name to which message data is published
            message: payload message (as bytes) to send.
            qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to `default_qos`
                config parameter or QOS_0.
            retain: retain flag. Defaults to ``default_retain`` config parameter or False.
            ack_timeout: duration to wait for connection acknowledgment from broker.

        Returns:
            the message that was sent

        """
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        def get_retain_and_qos() -> tuple[int, bool]:
            if qos is not None:
                if qos not in (QOS_0, QOS_1, QOS_2):
                    msg = f"QOS '{qos}' is not one of QOS_0, QOS_1, QOS_2."
                    raise ClientError(msg)
                _qos = qos
            else:
                _qos = self.config["default_qos"]
                with contextlib.suppress(KeyError):
                    _qos = self.config["topics"][topic]["qos"]
            if retain:
                _retain = retain
            else:
                _retain = self.config["default_retain"]
                with contextlib.suppress(KeyError):
                    _retain = self.config["topics"][topic]["retain"]
            return _qos, _retain

        (app_qos, app_retain) = get_retain_and_qos()
        return await self._handler.mqtt_publish(
            topic,
            message,
            app_qos,
            app_retain,
            ack_timeout,
        )

    @mqtt_connected
    async def subscribe(self, topics: list[tuple[str, int]]) -> list[int]:
        """Subscribe to topics.

        Send a MQTT [SUBSCRIBE](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718063)
        message and wait for broker acknowledgment.

        Args:
            topics: array of tuples containing topic pattern and QOS from `amqtt.mqtt.constants` to subscribe. For example:
                ```python
                [
                    ("$SYS/broker/uptime", QOS_1),
                    ("$SYS/broker/load/#", QOS_2),
                ]
                ```

        Returns:
            [SUBACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718068) message return code.

        """
        if self._handler and self.session:
            return await self._handler.mqtt_subscribe(topics, self.session.next_packet_id)
        return [0x80]

    @mqtt_connected
    async def unsubscribe(self, topics: list[str]) -> None:
        """Unsubscribe from topics.

        Send a MQTT [UNSUBSCRIBE](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718072)
        message and wait for broker [UNSUBACK](http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718077)
        message.

        Args:
            topics: array of topics to unsubscribe from.
                ```
                ["$SYS/broker/uptime", "$SYS/broker/load/#"]
                ```

        """
        if self._handler and self.session:
            await self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    async def deliver_message(self, timeout_duration: float | None = None) -> ApplicationMessage | None:
        """Deliver the next received message.

        Deliver next message received from the broker. If no message is available, this methods waits until next message arrives
        or ``timeout_duration`` occurs.

        Args:
            timeout_duration: maximum number of seconds to wait before returning. If not specified or None, there is no limit.

        Returns:
            instance of `amqtt.session.ApplicationMessage` containing received message information flow.

        Raises:
            asyncio.TimeoutError: if timeout occurs before a message is delivered
            ClientError: if client is not connected

        """
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        deliver_task = asyncio.create_task(self._handler.mqtt_deliver_next_message())
        self.client_tasks.append(deliver_task)
        self.logger.debug("Waiting for message delivery")

        done, _ = await asyncio.wait(
            [deliver_task],
            return_when=asyncio.FIRST_EXCEPTION,
            timeout=timeout_duration,
        )

        if self.client_tasks:
            self.client_tasks.pop()

        if deliver_task in done:
            exception = deliver_task.exception()
            if exception is not None:
                # deliver_task raised an exception, pass it on to our caller
                raise exception
            return deliver_task.result()
        # timeout occurred before message received
        deliver_task.cancel()
        msg = "Timeout waiting for message"
        raise asyncio.TimeoutError(msg)

    async def _connect_coro(self) -> int:
        """Perform the core connection logic."""
        if self.session is None:
            msg = "Session is not initialized."
            raise ClientError(msg)

        kwargs: dict[str, Any] = {}

        # Decode URI attributes
        uri_attributes = urlparse(self.session.broker_uri)
        scheme = uri_attributes.scheme
        secure = scheme in ("mqtts", "wss")
        self.session.username = (
            self.session.username or (str(uri_attributes.username) if uri_attributes.username else None)
        )
        self.session.password = (
            self.session.password or (str(uri_attributes.password) if uri_attributes.password else None)
        )
        self.session.remote_address = str(uri_attributes.hostname) if uri_attributes.hostname else None
        self.session.remote_port = uri_attributes.port

        if scheme in ("mqtt", "mqtts") and not self.session.remote_port:
            self.session.remote_port = 8883 if scheme == "mqtts" else 1883

        if scheme in ("ws", "wss") and not self.session.remote_port:
            self.session.remote_port = 443 if scheme == "wss" else 80

        if scheme in ("ws", "wss"):
            # Rewrite URI to conform to https://tools.ietf.org/html/rfc6455#section-3
            uri = (
                str(scheme),
                f"{self.session.remote_address}:{self.session.remote_port}",
                str(uri_attributes.path),
                str(uri_attributes.params),
                str(uri_attributes.query),
                str(uri_attributes.fragment),
            )
            self.session.broker_uri = str(urlunparse(uri))
        # Init protocol handler
        # if not self._handler:
        self._handler = ClientProtocolHandler(self.plugins_manager)

        connection_timeout = self.config.get("connection_timeout", None)

        if secure:
            sc = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH,
                cafile=self.session.cafile
            )

            if self.config.connection.certfile and self.config.connection.keyfile:
                sc.load_cert_chain(certfile=self.config.connection.certfile, keyfile=self.config.connection.keyfile)
            if self.config.connection.cafile:
                sc.load_verify_locations(cafile=self.config.connection.cafile)
            if self.config.check_hostname is not None:
                sc.check_hostname = self.config.check_hostname
                sc.verify_mode = ssl.CERT_REQUIRED
            kwargs["ssl"] = sc

        try:
            reader: StreamReaderAdapter | WebSocketsReader | None = None
            writer: StreamWriterAdapter | WebSocketsWriter | None = None
            self._connected_state.clear()

            # Open connection
            if scheme in ("mqtt", "mqtts"):
                conn_reader, conn_writer = await asyncio.wait_for(
                    asyncio.open_connection(
                        self.session.remote_address,
                        self.session.remote_port,
                        **kwargs,
                    ), timeout=connection_timeout)

                reader = StreamReaderAdapter(conn_reader)
                writer = StreamWriterAdapter(conn_writer)
            elif scheme in ("ws", "wss") and self.session.broker_uri:
                websocket: ClientConnection = await asyncio.wait_for(
                    websockets.connect(
                        self.session.broker_uri,
                        subprotocols=[websockets.Subprotocol("mqtt")],
                        additional_headers=self.additional_headers,
                        **kwargs,
                    ), timeout=connection_timeout)

                reader = WebSocketsReader(websocket)
                writer = WebSocketsWriter(websocket)
            elif not self.session.broker_uri:
                msg = "missing broker uri"
                raise ClientError(msg)
            else:
                msg = f"incorrect scheme defined in uri: '{scheme!r}'"
                raise ClientError(msg)

            # Start MQTT protocol
            self._handler.attach(self.session, reader, writer)
            return_code: int | None = await self._handler.mqtt_connect()

            if return_code is not CONNECTION_ACCEPTED:
                self.session.transitions.disconnect()
                self.logger.warning(f"Connection rejected with code '{return_code}'")
                msg = "Connection rejected by broker"
                exc = ConnectError(msg)
                exc.return_code = return_code
                raise exc
            # Handle MQTT protocol
            await self._handler.start()
            self.session.transitions.connect()
            self._connected_state.set()
            self.logger.debug(f"Connected to {self.session.remote_address}:{self.session.remote_port}")

        except (InvalidURI, InvalidHandshake, ProtocolHandlerError, ConnectionError, OSError, asyncio.TimeoutError) as e:
            self.logger.debug(f"Connection failed : {self.session.broker_uri} [{e!r}]")
            self.session.transitions.disconnect()
            raise ConnectError(e) from e
        return return_code

    async def handle_connection_close(self) -> None:
        """Handle disconnection from the broker."""
        if self.session is None:
            msg = "Session is not initialized."
            raise ClientError(msg)
        if self._handler is None:
            msg = "Handler is not initialized."
            raise ClientError(msg)

        def cancel_tasks() -> None:
            self._no_more_connections.set()
            while self.client_tasks:
                task = self.client_tasks.popleft()
                if not task.done():
                    task.cancel(msg="Connection closed.")

        self.logger.debug("Monitoring broker disconnection")
        # Wait for disconnection from broker (like connection lost)
        await self._handler.wait_disconnect()
        self.logger.warning("Disconnected from broker")

        # Block client API
        self._connected_state.clear()

        # stop an clean handler
        await self._handler.stop()
        self._handler.detach()
        self.session.transitions.disconnect()

        if self.config.get("auto_reconnect", False):
            # Try reconnection
            self.logger.debug("Auto-reconnecting")
            try:
                await self.reconnect()
            except ConnectError:
                # Cancel client pending tasks
                cancel_tasks()
        else:
            # Cancel client pending tasks
            cancel_tasks()

    def _init_session(
        self,
        uri: str | None = None,
        cleansession: bool | None = None,
        cafile: str | None = None,
        capath: str | None = None,
        cadata: str | None = None,
    ) -> Session:
        """Initialize the MQTT session."""
        broker_conf = self.config.get("connection", {}).copy()

        if uri is not None:
            broker_conf.uri = uri
        if cleansession is not None:
            self.config.cleansession = cleansession
        if cafile is not None:
            broker_conf.cafile = cafile
        if capath is not None:
            broker_conf.capath = capath
        if cadata is not None:
            broker_conf.cadata = cadata

        if not broker_conf.get("uri"):
            msg = "Missing connection parameter 'uri'"
            raise ClientError(msg)

        session = Session()
        session.broker_uri = broker_conf["uri"]
        session.client_id = self.client_id

        session.cafile = broker_conf.get("cafile")
        session.capath = broker_conf.get("capath")
        session.cadata = broker_conf.get("cadata")

        session.clean_session = self.config.get("cleansession", True)

        session.keep_alive = self.config["keep_alive"] - self.config["ping_delay"]

        if "will" in self.config:
            session.will_flag = True
            session.will_retain = self.config["will"]["retain"]
            session.will_topic = self.config["will"]["topic"]
            session.will_message = self.config["will"]["message"].encode()
            session.will_qos = self.config["will"]["qos"]

        return session



import asyncio
from asyncio import CancelledError
import logging

from amqtt.broker import Broker

"""
This sample shows how to run a broker
"""

formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
logging.basicConfig(level=logging.INFO, format=formatter)


async def run_server() -> None:
    broker = Broker()
    try:
        await broker.start()
        while True:
            await asyncio.sleep(1)
    except CancelledError:
        await broker.shutdown()

def __main__():
    try:
        asyncio.run(run_server())
    except KeyboardInterrupt:
        print("Server exiting...")

if __name__ == "__main__":
    __main__()



from abc import ABC, abstractmethod
from asyncio import StreamReader, StreamWriter
from contextlib import suppress
import io
import logging
import ssl
from typing import cast

from websockets import ConnectionClosed
from websockets.asyncio.connection import Connection


class ReaderAdapter(ABC):
    """Base class for all network protocol reader adapters.

    Reader adapters are used to adapt read operations on the network depending on the
    protocol used.
    """

    @abstractmethod
    async def read(self, n: int = -1) -> bytes:
        """Read up to n bytes. If n is not provided, or set to -1, read until EOF and return all read bytes.

        If the EOF was received and the internal buffer is
        empty, return an empty bytes object. :return: packet read as bytes data.
        """
        raise NotImplementedError

    @abstractmethod
    def feed_eof(self) -> None:
        """Acknowledge EOF."""
        raise NotImplementedError


class WriterAdapter(ABC):
    """Base class for all network protocol writer adapters.

    Writer adapters are used to adapt write operations on the network depending on
    the protocol used.
    """

    @abstractmethod
    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        raise NotImplementedError

    @abstractmethod
    async def drain(self) -> None:
        """Let the write buffer of the underlying transport a chance to be flushed."""
        raise NotImplementedError

    @abstractmethod
    def get_peer_info(self) -> tuple[str, int] | None:
        """Return peer socket info (remote address and remote port as tuple)."""
        raise NotImplementedError

    @abstractmethod
    def get_ssl_info(self) -> ssl.SSLObject | None:
        """Return peer certificate information (if available) used to establish a TLS session."""
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Close the protocol connection."""
        raise NotImplementedError


class WebSocketsReader(ReaderAdapter):
    """WebSockets API reader adapter.

    This adapter relies on Connection to read from a WebSocket.
    """

    def __init__(self, protocol: Connection) -> None:
        self._protocol = protocol
        self._stream = io.BytesIO(b"")

    async def read(self, n: int = -1) -> bytes:
        await self._feed_buffer(n)
        return self._stream.read(n)

    async def _feed_buffer(self, n: int = 1) -> None:
        """Feed the data buffer by reading a WebSocket message.

        :param n: Optional; feed buffer until it contains at least n bytes. Defaults to 1.
        """
        buffer = bytearray(self._stream.read())
        message: str | bytes | None = None
        while len(buffer) < n:
            with suppress(ConnectionClosed):
                message = await self._protocol.recv()
            if message is None:
                break
            message = message.encode("utf-8") if isinstance(message, str) else message
            buffer.extend(message)
        self._stream = io.BytesIO(buffer)

    def feed_eof(self) -> None:
        # NOTE: not implemented?!
        pass


class WebSocketsWriter(WriterAdapter):
    """WebSockets API writer adapter.

    This adapter relies on Connection to write to a WebSocket.
    """

    def __init__(self, protocol: Connection) -> None:
        self._protocol = protocol
        self._stream = io.BytesIO(b"")

    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        self._stream.write(data)

    async def drain(self) -> None:
        """Let the write buffer of the underlying transport a chance to be flushed."""
        data = self._stream.getvalue()
        if data and len(data):
            await self._protocol.send(data)
        self._stream = io.BytesIO(b"")

    def get_peer_info(self) -> tuple[str, int] | None:
        # remote_address can be either a 4-tuple or 2-tuple depending on whether
        # it is an IPv6 or IPv4 address, so we take their shared (host, port)
        # prefix here to present a uniform return value.
        remote_address: tuple[str, int] | None = self._protocol.remote_address[:2]
        return remote_address

    def get_ssl_info(self) -> ssl.SSLObject | None:
        return cast("ssl.SSLObject", self._protocol.transport.get_extra_info("ssl_object"))

    async def close(self) -> None:
        await self._protocol.close()


class StreamReaderAdapter(ReaderAdapter):
    """Asyncio Streams API protocol adapter.

    This adapter relies on StreamReader to read from a TCP socket.
    Because API is very close, this class is trivial.
    """

    def __init__(self, reader: StreamReader) -> None:
        self._reader = reader

    async def read(self, n: int = -1) -> bytes:
        if n == -1:
            data = await self._reader.read(n)
        else:
            data = await self._reader.readexactly(n)
        return data

    def feed_eof(self) -> None:
        self._reader.feed_eof()


class StreamWriterAdapter(WriterAdapter):
    """Asyncio Streams API protocol adapter.

    This adapter relies on StreamWriter to write to a TCP socket.
    Because API is very close, this class is trivial.
    """

    def __init__(self, writer: StreamWriter) -> None:
        self.logger = logging.getLogger(__name__)
        self._writer = writer
        self.is_closed = False  # StreamWriter has no test for closed...we use our own

    def write(self, data: bytes) -> None:
        if not self.is_closed:
            self._writer.write(data)

    async def drain(self) -> None:
        if not self.is_closed:
            await self._writer.drain()

    def get_peer_info(self) -> tuple[str, int]:
        extra_info = self._writer.get_extra_info("peername")
        return extra_info[0], extra_info[1]

    def get_ssl_info(self) -> ssl.SSLObject | None:
        return cast("ssl.SSLObject", self._writer.get_extra_info("ssl_object"))

    async def close(self) -> None:
        if not self.is_closed:
            self.is_closed = True  # we first mark this closed so yields below don't cause races with waiting writes
            await self._writer.drain()
            if self._writer.can_write_eof():
                self._writer.write_eof()
            self._writer.close()
            with suppress(AttributeError):
                await self._writer.wait_closed()


class BufferReader(ReaderAdapter):
    """Byte Buffer reader adapter.

    This adapter simply adapts reading a byte buffer.
    """

    def __init__(self, buffer: bytes) -> None:
        self._stream = io.BytesIO(buffer)

    async def read(self, n: int = -1) -> bytes:
        return self._stream.read(n)

    def feed_eof(self) -> None:
        # NOTE: not implemented?!
        pass


class BufferWriter(WriterAdapter):
    """ByteBuffer writer adapter.

    This adapter simply adapts writing to a byte buffer.
    """

    def get_ssl_info(self) -> ssl.SSLObject | None:
        return None

    def __init__(self, buffer: bytes = b"") -> None:
        self._stream = io.BytesIO(buffer)

    def write(self, data: bytes) -> None:
        """Write some data to the protocol layer."""
        self._stream.write(data)

    async def drain(self) -> None:
        pass

    def get_buffer(self) -> bytes:
        return self._stream.getvalue()

    def get_peer_info(self) -> tuple[str, int]:
        return "BufferWriter", 0

    async def close(self) -> None:
        self._stream.close()



from dataclasses import dataclass
from datetime import datetime, timedelta

try:
    from datetime import UTC
except ImportError:
    # support for python 3.10
    from datetime import timezone
    UTC = timezone.utc


from ipaddress import IPv4Address
import logging
from pathlib import Path
import re

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509 import Certificate, CertificateSigningRequest
from cryptography.x509.oid import NameOID

from amqtt.plugins.base import BaseAuthPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


class UserAuthCertPlugin(BaseAuthPlugin):
    """Used a *signed* x509 certificate's `Subject AlternativeName` or `SAN` to verify client authentication.

    Often used for IoT devices, this method provides the most secure form of identification. A root
    certificate, often referenced as a CA certificate -- either issued by a known authority (such as LetsEncrypt)
    or a self-signed certificate) is used to sign a private key and certificate for the server. Each device/client
    also gets a unique private key and certificate signed by the same CA certificate; also included in the device
    certificate is a 'SAN' or SubjectAlternativeName which is the device's unique identifier.

    Since both server and device certificates are signed by the same CA certificate, the client can
    verify the server's authenticity; and the server can verify the client's authenticity. And since
    the device's certificate contains a x509 SAN, the server (with this plugin) can identify the device securely.

    !!! note "URI and Client ID configuration"
        `uri_domain` configuration must be set to the same uri used to generate the device credentials

        when a device is connecting with private key and certificate, the `client_id` must
        match the device id used to generate the device credentials.

    Available ore three scripts to help with the key generation and certificate signing: `ca_creds`, `server_creds`
    and `device_creds`.

    !!! note "Configuring broker & client for using Self-signed root CA"
        If using self-signed root credentials, the `cafile` configuration for both broker and client need to be
        configured with `cafile` set to the `ca.crt`.
    """

    async def authenticate(self, *, session: Session) -> bool | None:
        """Verify the client's session using the provided client's x509 certificate."""
        if not session.ssl_object:
            return False

        der_cert = session.ssl_object.getpeercert(binary_form=True)
        if der_cert:
            cert = x509.load_der_x509_certificate(der_cert, backend=default_backend())

            try:
                san = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName)
                uris = san.value.get_values_for_type(x509.UniformResourceIdentifier)

                if self.config.uri_domain not in uris[0]:
                    return False

                pattern = rf"^spiffe://{re.escape(self.config.uri_domain)}/device/([^/]+)$"
                match = re.match(pattern, uris[0])
                if not match:
                    return False

                return match.group(1) == session.client_id

            except x509.ExtensionNotFound:
                logger.warning("No SAN extension found.")

        return False

    @dataclass
    class Config:
        """Configuration for the CertificateAuthPlugin."""

        uri_domain: str
        """The domain that is expected as part of the device certificate's spiffe (e.g. test.amqtt.io)"""


def generate_root_creds(country: str, state: str, locality: str,
                          org_name: str, cn: str) -> tuple[rsa.RSAPrivateKey, Certificate]:
    """Generate CA key and certificate."""
    # generate private key for the server
    ca_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
    )
    # Create certificate subject and issuer (self-signed)
    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, country),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state),
        x509.NameAttribute(NameOID.LOCALITY_NAME, locality),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, org_name),
        x509.NameAttribute(NameOID.COMMON_NAME, cn),
    ])

    # 3. Build self-signed certificate
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(ca_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(UTC))
        .not_valid_after(datetime.now(UTC) + timedelta(days=3650))  # 10 years
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(ca_key.public_key()),
            critical=False,
        )
        .add_extension(
            x509.KeyUsage(
                key_cert_sign=True,
                crl_sign=True,
                digital_signature=False,
                key_encipherment=False,
                content_commitment=False,
                data_encipherment=False,
                key_agreement=False,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .sign(ca_key, hashes.SHA256())
    )

    return ca_key, cert


def generate_server_csr(country: str, org_name: str, cn: str) -> tuple[rsa.RSAPrivateKey, CertificateSigningRequest]:
    """Generate server private key and server certificate-signing-request."""
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, country),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, org_name),
            x509.NameAttribute(NameOID.COMMON_NAME, cn),
        ]))
        .add_extension(
            x509.SubjectAlternativeName([
                x509.DNSName(cn),
                x509.IPAddress(IPv4Address("127.0.0.1")),
            ]),
            critical=False,
        )
        .sign(key, hashes.SHA256())
    )

    return key, csr


def generate_device_csr(country: str, org_name: str, common_name: str,
                        uri_san: str, dns_san: str
                        ) -> tuple[rsa.RSAPrivateKey, CertificateSigningRequest]:
    """Generate a device key and a csr."""
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, country),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, org_name
                               ),
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]))
        .add_extension(
            x509.SubjectAlternativeName([
                x509.UniformResourceIdentifier(uri_san),
                x509.DNSName(dns_san),
            ]),
            critical=False,
        )
        .sign(key, hashes.SHA256())
    )

    return key, csr


def sign_csr(csr: CertificateSigningRequest,
             ca_key: rsa.RSAPrivateKey,
             ca_cert: Certificate, validity_days: int = 365) -> Certificate:
    """Sign a csr with CA credentials."""
    return (
        x509.CertificateBuilder()
        .subject_name(csr.subject)
        .issuer_name(ca_cert.subject)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(UTC))
        .not_valid_after(datetime.now(UTC) + timedelta(days=validity_days))
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        .add_extension(
            csr.extensions.get_extension_for_class(x509.SubjectAlternativeName).value,
            critical=False,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_cert.public_key()),  # type: ignore[arg-type]
            critical=False,
        )
        .sign(ca_key, hashes.SHA256())
    )


def load_ca(ca_key_fn: str, ca_crt_fn: str) -> tuple[rsa.RSAPrivateKey, Certificate]:
    """Load server key and certificate."""
    with Path(ca_key_fn).open("rb") as f:
        ca_key: rsa.RSAPrivateKey = serialization.load_pem_private_key(f.read(), password=None)  # type: ignore[assignment]
    with Path(ca_crt_fn).open("rb") as f:
        ca_cert = x509.load_pem_x509_certificate(f.read())
    return ca_key, ca_cert


def write_key_and_crt(key: rsa.RSAPrivateKey, crt: Certificate,
                      prefix: str, path: Path | None = None) -> None:
    """Create pem-encoded files for key and certificate."""
    path = path or Path()

    crt_fn = path / f"{prefix}.crt"
    key_fn = path / f"{prefix}.key"

    with crt_fn.open("wb") as f:
        f.write(crt.public_bytes(serialization.Encoding.PEM))
    with key_fn.open("wb") as f:
        f.write(key.private_bytes(
            serialization.Encoding.PEM,
            serialization.PrivateFormat.TraditionalOpenSSL,
            serialization.NoEncryption()
        ))
